

"Functions implementing css editing"


import collections

from skipole import ValidateError, FailPage, ServerError

from skipole. import skilift
from skipole.skilift import editpage

from .. import utils


def retrieve_edit_csspage(skicall):
    "Retrieves widget data for the edit css page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data, keeping page_number, pchange and any status message
    utils.clear_call_data(call_data, keep=["page_number", "pchange", "status"])

    project = call_data['editedprojname']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
        str_pagenumber = str(pagenumber)
    else:
        raise FailPage(message = "page missing")
    if not pagenumber:
        raise FailPage(message = "Invalid page")
    try:
        pageinfo = skilift.page_info(project, pagenumber)
        if pageinfo.item_type != 'CSS':
            raise FailPage(message = "Invalid page")

        call_data['pchange'] = pageinfo.change

        selectors = list(editpage.css_style(project, pagenumber).keys())
    except ServerError as e:
        raise FailPage(message = e.message)

    # fills in the data for editing page name, brief, parent, etc., 
    page_data[("adminhead","page_head","large_text")] = pageinfo.name
    page_data[('page_edit','p_ident','page_ident')] = (project,str_pagenumber)
    page_data[('page_edit','p_name','page_ident')] = (project,str_pagenumber)
    page_data[('page_edit','p_description','page_ident')] = (project,str_pagenumber)
    page_data[('page_edit','p_rename','input_text')] = pageinfo.name
    page_data[('page_edit','p_parent','input_text')] = "%s,%s" % (project, pageinfo.parentfolder_number)
    page_data[('page_edit','p_brief','input_text')] = pageinfo.brief
    # create the contents for the selectortable
    contents = []

    if selectors:
        max_selector_index = len(selectors) - 1
        for index,selector in enumerate(selectors):
            if index:
                up = True
            else:
                # first item (index zero) has no up button
                up = False
            if index < max_selector_index:
                down = True
            else:
                # last item has no down button
                down = False 
            contents.append([selector, selector, selector, selector, selector, True, up, down, True])
        page_data["selectortable:contents"] = contents
    else:
        page_data["selectortable:show"] = False

    page_data['enable_cache:radio_checked'] = pageinfo.enable_cache


def retrieve_edit_selector(skicall):
    "Retrieves widget data for the edit selector page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
        str_pagenumber = str(pagenumber)
    else:
        raise FailPage(message = "page missing")
    if not pagenumber:
        raise FailPage(message = "Invalid page")
    if 'edit_selector' not in call_data:
        raise FailPage(message="No selector given")

    try:
        pageinfo = skilift.page_info(project, pagenumber)
        if pageinfo.item_type != 'CSS':
            raise FailPage(message = "Invalid page")
        edit_selector = call_data['edit_selector']
        property_string = editpage.css_selector_properties(project, pagenumber, edit_selector)
    except ServerError as e:
        raise FailPage(message = e.message)

    call_data['extend_nav_buttons'] = [['back_to_css_page', 'CSS Page', True, '']]    # label to 8003

    if 'status' in call_data:
        status_text = call_data['status']
    else:
        status_text = 'Edit selector : %s' % (edit_selector,)

    page_data.update({("adminhead","page_head","large_text"):"Edit CSS page : %s" % (pageinfo.name,),
                      ("adminhead","page_head","small_text"):status_text,
                      ('selectorname','para_text'):"Selector : %s" % (edit_selector,),
                      ('p_ident','page_ident'):(project,str_pagenumber),
                      ('properties','hidden_field1'):edit_selector,
                      ('properties','input_text'):property_string })



def retrieve_print_csspage(skicall):
    "Retrieves widget data for the print css page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message = "page missing")
    if not pagenumber:
        raise FailPage(message = "Invalid page")
    try:
        pageinfo = skilift.item_info(project, pagenumber)
        if pageinfo.item_type != 'CSS':
            raise FailPage(message = "Invalid page")
        page_string = editpage.page_string(project, pagenumber)
    except ServerError as e:
        raise FailPage(message = e.message)
    page_data['page_details:para_text'] = "/**************************\n\nIdent:%s,%s\nPath:%s\n%s" % (project, pagenumber, pageinfo.path, pageinfo.brief)
    page_data['page_contents:para_text'] = page_string


def submit_new_selector(skicall):
    "Sets new selector"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message = "page missing")
    if not pagenumber:
        raise FailPage(message = "Invalid page")
    pchange = call_data['pchange']
    if not 'new_selector' in call_data:
        raise FailPage(message="No new selector given")
    new_selector = call_data['new_selector']
    if not new_selector:
        raise FailPage(message="No selector given")
    try:
        style = editpage.css_style(project, pagenumber)
        if new_selector in style:
            raise FailPage(message="Selector already exists")
        # Set the selector
        style[new_selector] = []
        call_data['pchange'] = editpage.set_css_style(project, pagenumber, pchange, style)
    except ServerError as e:
        raise FailPage(message=e.message)
    call_data['status'] = 'Selector \"%s\" added.' % (new_selector,)


def submit_delete_selector(skicall):
    "Deletes selector"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message = "page missing")
    if not pagenumber:
        raise FailPage(message = "Invalid page")
    pchange = call_data['pchange']
    if 'remove_selector' not in call_data:
        raise FailPage(message="No selector to delete given")
    remove_selector = call_data['remove_selector']
    if not remove_selector:
        raise FailPage(message="No selector given")
    try:
        style = editpage.css_style(project, pagenumber)
        if remove_selector not in style:
            raise FailPage(message="Selector to remove is not present")
        # Remove the selector
        del style[remove_selector]
        call_data['pchange'] = editpage.set_css_style(project, pagenumber, pchange, style)
    except ServerError as e:
        raise FailPage(message=e.message)
    call_data['status'] = 'Selector deleted'


def submit_selector_properties(skicall):
    "Sets the selector properties"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message = "page missing")
    if not pagenumber:
        raise FailPage(message = "Invalid page")
    pchange = call_data['pchange']
    if 'edit_selector' not in call_data:
        raise FailPage(message="No selector to edit given")
    edit_selector = call_data['edit_selector']
    if 'properties' not in call_data:
        raise FailPage(message="No properties to set given")
    property_string = call_data['properties']
    property_list = property_string.split(';')
    properties = []
    for item in property_list:
        if not item:
            continue
        if ':' not in item:
            continue
        pair = item.split(':')
        if len(pair) != 2:
            raise FailPage(message="Invalid properties")
        a = pair[0].strip()
        b = pair[1].strip()
        properties.append([a,b])
    if not properties:
        raise FailPage(message="Invalid properties")
    try:
        style = editpage.css_style(project, pagenumber)
        style[edit_selector] = properties
        call_data['pchange'] = editpage.set_css_style(project, pagenumber, pchange, style)
    except ServerError as e:
        raise FailPage(message=e.message)
    call_data['status'] = 'Properties set'
        

def move_selector_up(skicall):
    "Moves selector upwards"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message = "page missing")
    if not pagenumber:
        raise FailPage(message = "Invalid page")
    pchange = call_data['pchange']
    if 'up_selector' not in call_data:
        raise FailPage(message="No selector to move up given")
    up_selector = call_data['up_selector']
    if not up_selector:
        raise FailPage(message="No selector given")
    try:
        style = editpage.css_style(project, pagenumber)
        if up_selector not in style:
            raise FailPage(message="Selector to move up is not present")
        selector_list = list(style.keys())
        # Move the selector up
        idx = selector_list.index(up_selector)
        if not idx:
            # idx cannot be zero
            raise FailPage(message="Selector already at top")
        selector_list[idx-1], selector_list[idx] = selector_list[idx], selector_list[idx-1]
        new_style = collections.OrderedDict([(selector,style[selector]) for selector in selector_list])
        call_data['pchange'] = editpage.set_css_style(project, pagenumber, pchange, new_style)
    except ServerError as e:
        raise FailPage(message=e.message)


def move_selector_down(skicall):
    "Moves selector downwardswards"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message = "page missing")
    if not pagenumber:
        raise FailPage(message = "Invalid page")
    pchange = call_data['pchange']
    if 'down_selector' not in call_data:
        raise FailPage(message="No selector to move down given")
    down_selector = call_data['down_selector']
    if not down_selector:
        raise FailPage(message="No selector given")
    try:
        style = editpage.css_style(project, pagenumber)
        if down_selector not in style:
            raise FailPage(message="Selector to move down is not present")
        # Move the selector down
        selector_list = list(style.keys())
        idx = selector_list.index(down_selector)
        if idx == len(selector_list) -1:
            # idx cannot be last element
            raise FailPage(message="Selector already at end")
        selector_list[idx+1], selector_list[idx] = selector_list[idx], selector_list[idx+1]
        new_style = collections.OrderedDict([(selector,style[selector]) for selector in selector_list])
        call_data['pchange'] = editpage.set_css_style(project, pagenumber, pchange, new_style)
    except ServerError as e:
        raise FailPage(message=e.message)


def submit_cache(skicall):
    "Sets cache true or false"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if 'cache' not in call_data:
        raise FailPage(message="No cache instruction given")
    try:
        # Set the page cache
        if call_data['cache'] == 'True':
            enable_cache = True
            message = "Cache Enabled"
        else:
            enable_cache = False
            message = "Cache Disabled"
        call_data['pchange'] = editpage.page_enable_cache(project, pagenumber, pchange, enable_cache)
    except ServerError as e:
        raise FailPage(message=e.message)
    call_data['status'] = message






"""Contains widgets for inputting data"""

from urllib.parse import quote

from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict




class InputTable1(Widget):
    """A table of three columns, two being text strings, the third a text input field
       The first row is three header titles. Typically inserted into a form
       does not display errors"""

    display_errors = False


    arg_descriptions = {'header_class':FieldArg("cssclass",""),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'title1':FieldArg('text', ''),
                        'title2':FieldArg('text', ''),
                        'title3':FieldArg('text', ''),
                        'row_classes':FieldArgList('text', jsonset=True),
                        'col1':FieldArgList('text', jsonset=True),
                        'col2':FieldArgList('text', jsonset=True),
                        'inputdict':FieldArgDict('text', valdt=True, jsonset=True, senddict=True)                 # dictionary of keyname:value
                        }                                                                                         # the field submitted as a dictionary
                                                                                                                  # each key will have 'keyname'
                                                                                                                  # and value will be the values ticked

    def __init__(self, name=None, brief='', **field_args):
        """
        header_class: class of the header row, if empty string, then no class will be applied
        size: The number of characters appearing in each text input area
        maxlength: The maximum number of characters accepted in each text area
        title1: The header title over the first text column
        title2: The header title over the second text column
        title3: The header title over the third text column
        row_classes: A list of CSS classes to apply to each row (not including the header)
        col1: A list of text strings to place in the first column
        col2: A list of text strings to place in the second column
        inputdict: A dictionary of keyname:value, should be Ordered Dict unless python >= 3.6
                   the fields submitted will have name 'widgetname:inputdict-keyname'
                   and the user will receive a widgfield ('widgetname','inputdict') containing a dictionary of keyname:values
         """
        Widget.__init__(self, name=name, tag_name="table", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the table"
        inputdict = self.get_field_value("inputdict")
        rowc = self.get_field_value("row_classes")
        col1 = self.get_field_value("col1")
        col2 = self.get_field_value("col2")
        input_name = self.get_formname("inputdict") + '-'
        size = self.get_field_value('size')
        maxlength = self.get_field_value('maxlength')

        header = 0
        if self.get_field_value('title1') or self.get_field_value('title2') or self.get_field_value('title3'):
            header = 1
            if self.get_field_value('header_class'):
                self[0] = tag.Part(tag_name='tr', attribs={"class":self.get_field_value('header_class')})
            else:
                self[0] = tag.Part(tag_name='tr')
            self[0][0] = tag.Part(tag_name='th', text = self.get_field_value('title1'))
            self[0][1] = tag.Part(tag_name='th', text = self.get_field_value('title2'))
            self[0][2] = tag.Part(tag_name='th', text = self.get_field_value('title3'))

        # create rows
        rows = max( len(col1), len(col2), len(inputdict) )
        if not rowc:
            rowc = ['']*rows
        if rows > len(rowc):
            rowc.extend(['']*(rows - len(rowc)))
        if rows > len(col1):
            col1.extend(['']*(rows - len(col1)))
        if rows > len(col2):
            col2.extend(['']*(rows - len(col2)))

        keylist = list(inputdict.keys())
        if rows > len(keylist):
            keylist.extend([None]*(rows - len(keylist)))

        # keylist is a list of the dictionary keys, extended by None keys, if the dictionary is smaller than the number of rows of the table
        
        for index in range(rows):
            rownumber = index+header
            if rowc[index]:
                self[rownumber] = tag.Part(tag_name='tr', attribs={"class":rowc[index]})
            else:
                self[rownumber] = tag.Part(tag_name='tr')
            self[rownumber][0] = tag.Part(tag_name='td', text = col1[index])
            self[rownumber][1] = tag.Part(tag_name='td', text = col2[index])
            self[rownumber][2] = tag.Part(tag_name='td')
            key = keylist[index]
            if key:  # this is the dictionary key
                keyed_name = input_name + key
                self[rownumber][2][0] = tag.ClosedPart(tag_name="input", attribs={"name":keyed_name, "type":"text", "value":inputdict[key]})
                if size:
                    self[rownumber][2][0].update_attribs({"size":size})
                if maxlength:
                    self[rownumber][2][0].update_attribs({"maxlength":maxlength})



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<table>  <!-- with widget id and class widget_class -->
  <tr> <!-- with header class -->
    <th> <!-- title1 --> </th>
    <th> <!-- title2 --> </th>
    <th> <!-- title3 --> </th>
  </tr>
  <tr> <!-- with class from row_classes -->
    <td> <!-- col1 text string --> </td>
    <td> <!-- col2 text string --> </td>
    <td><input type="text" />  <!-- with names and value derived from 'inputdict' -->
    </td>
  </tr>
  <!-- rows repeated -->
</table>"""





class InputTable5(Widget):
    """Defines a div of four columns, the first just being label text, the second being text input fields,
       the third and fourth being submit buttons.  Each row of the table is a form, the
       form action ident is the same for all forms.
       On error the paragraph TextBlock is changed to the error message."""

    # This class does not display any error messages
    display_errors = False

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {
                        'div_class':FieldArg("cssclass", ''),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'required':FieldArg("boolean", False),
                        'action':FieldArg("url",''),
                        'button_text1':FieldArg("text", "Submit", valdt=True),
                        'button_text2':FieldArg("text", "Submit", valdt=True),
                        'button1_class':FieldArg("cssclass", ""),
                        'button2_class':FieldArg("cssclass", ""),
                        'col_label':FieldArgList('text'),
                        'label_class':FieldArg("cssclass", ""),
                        'label_style':FieldArg("cssstyle", ""),
                        'col_input':FieldArgList('text', valdt=True),
                        'hidden_field1':FieldArgList("text", valdt=True),
                        'hidden_field2':FieldArgList("text", valdt=True),
                        'hidden_field3':FieldArgList("text", valdt=True),
                        'hidden_field4':FieldArgList("text", valdt=True),
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_class':FieldArg("cssclass", ''),
                        'input_style':FieldArg("cssstyle", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArgDict("boolean", jsonset=True),
                        'set_input_errored':FieldArgDict("boolean", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        div_class: the class attribute of each form div
        size: The number of characters appearing in each text input area
        maxlength: The maximum number of characters accepted in each text area
        required: Set True to put the 'required' flag into each input field
        action: The label or ident of the action page
        button_text1: text appearing on the first buttons - if empty, no button 1 is shown
        button_text2: text appearing on the second buttons - if empty, no button 2 is shown
        button1_class: class set on button1
        button2_class: class set on button2
        col_label: a list of all the text strings appearing in the first column
        label_class: the class to apply to each label
        col_input: a list of all the text strings appearing in the input fields
        hidden_field1-4: four hidden fields, each being a list
        input_accepted_class: A class which can be added on each input field
        input_class: A class which can be set on each input field
        input_errored_class: A class which can be added on each input field
        The following two dictionaries should have keys equal to the integer row number (starting
         at zero) and value True or False
        set_input_accepted: Dictionary of col index keys that should be set with input_accepted_class
        set_input_errored: Dictionary of col index keys that should be set with input_errored_class
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the table"

        # get the widget ident
        ident_value = page.ident_data_string

        # get the form action url
        action_url = skiboot.get_url(self.get_field_value("action"), proj_ident=page.proj_ident)
        if not action_url:
            # setting self._error replaces the entire tag by the self._error message
            self._error = "Warning: broken link"
            return


        len_label = len(self.fields["col_label"])
        len_input = len(self.fields["col_input"])
        len_hidden_field1 = len(self.fields["hidden_field1"])
        len_hidden_field2 = len(self.fields["hidden_field2"])
        len_hidden_field3 = len(self.fields["hidden_field3"])
        len_hidden_field4 = len(self.fields["hidden_field4"])

        rows = max(len_label, len_input)
        if not rows:
            rows=1

        input_name = self.get_formname("col_input")
        button_name1 = self.get_formname("button_text1")
        button_value1 = self.get_field_value('button_text1')
        button1_class = self.get_field_value('button1_class')
        button_name2 = self.get_formname("button_text2")
        button_value2 = self.get_field_value('button_text2')
        button2_class = self.get_field_value('button2_class')
        hidden_field1_name = self.get_formname("hidden_field1")
        hidden_field2_name = self.get_formname("hidden_field2")
        hidden_field3_name = self.get_formname("hidden_field3")
        hidden_field4_name = self.get_formname("hidden_field4")

        label_class = self.get_field_value('label_class')
        label_style = self.get_field_value('label_style')

        form_id = self.get_id() + '_'

        if self.get_field_value('input_accepted_class') and self.get_field_value("set_input_accepted"):
            set_input_accepted = self.get_field_value("set_input_accepted")
        else:
            set_input_accepted = {}

        if self.get_field_value('input_errored_class') and self.get_field_value("set_input_errored"):
            set_input_errored = self.get_field_value("set_input_errored")
        else:
            set_input_errored = {}

        input_class = self.get_field_value('input_class')
        input_style = self.get_field_value('input_style')


        for rownumber in range(rows):
            if rownumber<len_label:
                col_label = self.get_field_value("col_label")[rownumber]
            else:
                col_label = ''
            if rownumber<len_input:
                col_input = self.get_field_value("col_input")[rownumber]
            else:
                col_input = ''
            if rownumber<len_hidden_field1:
                col_hidden_field1 = self.get_field_value("hidden_field1")[rownumber]
            else:
                col_hidden_field1 = ''
            if rownumber<len_hidden_field2:
                col_hidden_field2 = self.get_field_value("hidden_field2")[rownumber]
            else:
                col_hidden_field2 = ''
            if rownumber<len_hidden_field3:
                col_hidden_field3 = self.get_field_value("hidden_field3")[rownumber]
            else:
                col_hidden_field3 = ''
            if rownumber<len_hidden_field4:
                col_hidden_field4 = self.get_field_value("hidden_field4")[rownumber]
            else:
                col_hidden_field4 = ''

            # for every table row, create a form
            formrow = tag.Part(tag_name="form", attribs ={"method":"post", "action": action_url})

            # 1st column is label text 
            if label_class:
                formrow[0] = tag.Part(tag_name="label", attribs={'class':label_class})
            else:
                formrow[0] = tag.Part(tag_name="label")
            if label_style:
                formrow[0].update_attribs({"style":label_style})
            formrow[0][0] = col_label

            # 2nd column is a text input field
            formrow[1] = tag.ClosedPart(tag_name="input",
                                    attribs ={"name":input_name,
                                    "value":col_input,
                                    "type":"text"})
            if self.get_field_value('size'):
                formrow[1].update_attribs({"size":self.get_field_value('size')})
            if self.get_field_value('maxlength'):
                formrow[1].update_attribs({"maxlength":self.get_field_value('maxlength')})
            if self.get_field_value('required'):
                formrow[1].update_attribs({"required":"required"})
            # set an id in the input field for the 'label for' tag
            this_id = form_id + str(rownumber) + '_0_1'
            formrow[1].insert_id(this_id)
            # set the label 'for' attribute
            formrow[0].update_attribs({'for':formrow[1].get_id()})
            if (rownumber in set_input_errored) and set_input_errored[rownumber]:
                if input_class:
                    formrow[1].update_attribs({"class":input_class + ' ' + self.get_field_value('input_errored_class')})
                else:
                    formrow[1].update_attribs({"class":self.get_field_value('input_errored_class')})
            elif (rownumber in set_input_accepted) and set_input_accepted[rownumber]:
                if input_class:
                    formrow[1].update_attribs({"class":input_class + ' ' + self.get_field_value('input_accepted_class')})
                else:
                    formrow[1].update_attribs({"class":self.get_field_value('input_accepted_class')})
            elif input_class:
                formrow[1].update_attribs({"class":input_class})

            if input_style:
                formrow[1].update_attribs({"style":input_style})

            # 3rd column is a submit button
            if button_value1:
                formrow.append(tag.ClosedPart(tag_name="input",
                                 attribs ={"name":button_name1,
                                           "value":button_value1,
                                           "class":button1_class,
                                           "type":"submit"}))
            # 4th column is a submit button
            if button_value2:
                formrow.append(tag.ClosedPart(tag_name="input",
                                 attribs ={"name":button_name2,
                                           "value":button_value2,
                                           "class":button2_class,
                                           "type":"submit"}))

            # all submissions always have an 'ident' hidden field to provide the ident of the calling page
            formrow.append(tag.ClosedPart(tag_name="input",
                                          attribs ={"name":'ident',
                                                 "value":ident_value,
                                                 "type":"hidden"}))

            # hidden field on the form
            if col_hidden_field1:
                formrow.append(tag.ClosedPart(tag_name = "input",
                                            attribs =  {"name":hidden_field1_name,
                                                        "value":col_hidden_field1,
                                                        "type":"hidden"}))

            # Second hidden field on the form
            if col_hidden_field2:
                formrow.append(tag.ClosedPart(tag_name = "input",
                                            attribs =  {"name":hidden_field2_name,
                                                        "value":col_hidden_field2,
                                                        "type":"hidden"}))
            # Third hidden field on the form
            if col_hidden_field3:
                formrow.append(tag.ClosedPart(tag_name = "input",
                                            attribs =  {"name":hidden_field3_name,
                                                        "value":col_hidden_field3,
                                                        "type":"hidden"}))
            # Fourth hidden field on the form
            if col_hidden_field4:
                formrow.append(tag.ClosedPart(tag_name = "input",
                                            attribs =  {"name":hidden_field4_name,
                                                        "value":col_hidden_field4,
                                                        "type":"hidden"}))

            # place form in a div
            self[rownumber] = tag.Part(tag_name="div")
            if self.get_field_value('div_class'):
                self[rownumber].attribs = {"class": self.get_field_value('div_class')}
            self[rownumber][0] = formrow


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        jscript = """  $('#{ident}').find( 'form').each(function() {{
    $(this).on("submit", function(e) {{
        SKIPOLE.widgets['{ident}'].eventfunc(e);
        }});
    }});
""".format(ident=self.get_id())
        return jscript + self._make_fieldvalues('input_accepted_class', 'input_errored_class')

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>   <!-- with widget id and class widget_class -->
  <!-- Then a div containing a form, label, input field, buttons and hidden fields, repeated for each list  item -->
  <div> <!-- with CSS class attribute set to div_class if a class is set -->
    <form method=\"post\"> <!-- action attribute set to action field -->
      <label> <!-- with col_label text and CSS class label_class -->
      </label >
        <input type="text" /> <!-- all with same CSS input_class and size and maxlength, with values from col_input -->
        <input type="submit" /> <!-- with value button_text1, CSS class button1_class and name widgfield derived from button_text1 field name -->
        <input type="submit" /> <!-- with value button_text2, CSS class button2_class and name widgfield derived from button_text2 field name -->
        <!-- hidden input fields -->
    </form>                              
  </div>
  <!--  The above repeated for each item in col_input or col_label, whichever is greater -->
</div>"""




class InputTable4(Widget):
    """A table of four columns, three being text strings,
       the fourth being a cell with up and down arrow links and a hidden field
       On the links being chosen a request for a json update is made, which
       typically updates the text columns, and the values in the hidden field.
       Each arrow has two get fields which can be set.
       The first row is four header titles.
       Typically inserted into a form, does not display errors"""

    display_errors = False


    arg_descriptions = {'header_class':FieldArg("cssclass",""),
                        'col1_class':FieldArg("cssclass",""),          # class applied to every td in the first column
                        'col2_class':FieldArg("cssclass",""),          # class applied to every td in the second column
                        'col3_class':FieldArg("cssclass",""),          # class applied to every td in the third column
                        'col4_class':FieldArg("cssclass",""),          # class applied to every td in the fourth column
                        'col4_style':FieldArg("cssstyle", ""),
                        'row_classes':FieldArgList('text', jsonset=True),
                        'title1':FieldArg('text', ''),
                        'title2':FieldArg('text', ''),
                        'title3':FieldArg('text', ''),
                        'title4':FieldArg('text', ''),
                        'up_link_ident':FieldArg("url", 'no_javascript'),
                        'up_json_ident':FieldArg("url", ''),
                        'down_link_ident':FieldArg("url", 'no_javascript'),
                        'down_json_ident':FieldArg("url", ''),
                        'up_style':FieldArg("cssstyle", ""),
                        'down_style':FieldArg("cssstyle", ""),
                        'up_class':FieldArg("cssclass", ""),
                        'down_class':FieldArg("cssclass", ""),
                        'up_hide':FieldArgList("boolean", jsonset=True),
                        'down_hide':FieldArgList("boolean", jsonset=True),
                        'up_getfield1':FieldArgList('text', valdt=True, jsonset=True),
                        'up_getfield2':FieldArgList('text', valdt=True, jsonset=True),
                        'down_getfield1':FieldArgList('text', valdt=True, jsonset=True),
                        'down_getfield2':FieldArgList('text', valdt=True, jsonset=True),
                        'col1':FieldArgList('text', jsonset=True),
                        'col2':FieldArgList('text', jsonset=True),
                        'col3':FieldArgList('text', jsonset=True),
                        'inputdict':FieldArgDict('text', valdt=True, jsonset=True, senddict=True)                 # dictionary of keyname:value
                        }                                                                                         # the field submitted as a dictionary
                                                                                                                  # each key will have 'keyname'
                                                                                                                  # and value will be the hidden field values

    def __init__(self, name=None, brief='', **field_args):
        """
        header_class: class of the header row, if empty string, then no class will be applied
        col1_class: class applied to every td in the first column
        col2_class: class applied to every td in the second column
        col3_class: class applied to every td in the third column
        col4_class: class applied to every td in the fourth column
        col4_style: style applied to every td in the fourth column
        row_classes: A list of CSS classes to apply to each row (not including the header)
        title1: The header title over the first text column
        title2: The header title over the second text column
        title3: The header title over the third text column
        title4: The header title over the fourth text column
        up_link_ident: ident of the up arrow link if javascript disabled
        up_json_ident: ident of the up arrow link, expects a json file returned
        down_link_ident: ident of the down arrow link if javascript disabled
        down_json_ident: ident of the down arrow link, expects a json file returned
        up_style: CSS style applied to the up arrows
        down_style: CSS style applied to the down arrows
        up_class: CSS class applied to the up arrows
        down_class: CSS class applied to the down arrows
        up_hide: List of True/False to hide or not hide the up arrow
        down_hide: List of True/False to hide or not hide the down arrow
        up_getfield1: list of get fields, one for each up arrow link
        up_getfield2: list of second get fields, one for each up arrow link
        down_getfield1: list of get fields, one for each down arrow link
        down_getfield2: list of get fields, one for each down arrow link
        col1: A list of text strings to place in the first column
        col2: A list of text strings to place in the second column
        col3: A list of text strings to place in the third column
        inputdict: A dictionary of keyname:value, should be Ordered Dict unless python >= 3.6
                   the fields submitted will have name 'widgetname:inputdict-keyname'
                   and the user will receive a widgfield ('widgetname','inputdict') containing a dictionary of keyname:values
         """
        Widget.__init__(self, name=name, tag_name="table", brief=brief, **field_args)
        self._up_jsonurl = ''
        self._down_jsonurl = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the table"
        inputdict = self.get_field_value("inputdict")
        rowc = self.get_field_value("row_classes")
        col1 = self.get_field_value("col1")
        col2 = self.get_field_value("col2")
        col3 = self.get_field_value("col3")
        up_hide = self.get_field_value("up_hide")
        down_hide = self.get_field_value("down_hide")


        input_name = self.get_formname("inputdict") + '-'

        col1_class = self.get_field_value("col1_class")
        col2_class = self.get_field_value("col2_class")
        col3_class = self.get_field_value("col3_class")
        col4_class = self.get_field_value("col4_class")
        col4_style = self.get_field_value("col4_style")

        up_link_ident = self.get_field_value("up_link_ident") # ident of the up arrow link if javascript disabled
        if not up_link_ident:
            up_link_ident = 'no_javascript'
        up_json_ident = self.get_field_value("up_json_ident") # ident of the up arrow link, expects a json file returned

        down_link_ident = self.get_field_value("down_link_ident") # ident of the down arrow link if javascript disabled
        if not down_link_ident:
            down_link_ident = 'no_javascript'
        down_json_ident = self.get_field_value("down_json_ident") # ident of the down arrow link, expects a json file returned

        up_style = self.get_field_value("up_style") # CSS style applied to the up arrows
        down_style = self.get_field_value("down_style") # CSS style applied to the down arrows
        up_class = self.get_field_value("up_class") # CSS class applied to the up arrows
        down_class = self.get_field_value("down_class") # CSS class applied to the down arrows

        up_getfield1 = self.get_field_value("up_getfield1") # list of get fields, one for each up arrow link
        up_getfield2 = self.get_field_value("up_getfield2") # list of second get fields, one for each up arrow link
        down_getfield1 = self.get_field_value("down_getfield1") # list of get fields, one for each down arrow link
        down_getfield2 = self.get_field_value("down_getfield2") # list of get fields, one for each down arrow link

        # up arrow link
        # get json url
        if up_json_ident:
            self._up_jsonurl = skiboot.get_url(up_json_ident, proj_ident=page.proj_ident)
        # get url
        up_url = skiboot.get_url(up_link_ident, proj_ident=page.proj_ident)

        # down arrow link
        # get json url
        if down_json_ident:
            self._down_jsonurl = skiboot.get_url(down_json_ident, proj_ident=page.proj_ident)
        # get url
        down_url = skiboot.get_url(down_link_ident, proj_ident=page.proj_ident)


        header = 0
        if self.get_field_value('title1') or self.get_field_value('title2') or self.get_field_value('title3') or self.get_field_value('title4'):
            header = 1
            if self.get_field_value('header_class'):
                self[0] = tag.Part(tag_name='tr', attribs={"class":self.get_field_value('header_class')})
            else:
                self[0] = tag.Part(tag_name='tr')
            self[0][0] = tag.Part(tag_name='th', text = self.get_field_value('title1'))
            self[0][1] = tag.Part(tag_name='th', text = self.get_field_value('title2'))
            self[0][2] = tag.Part(tag_name='th', text = self.get_field_value('title3'))
            self[0][3] = tag.Part(tag_name='th', text = self.get_field_value('title4'))

        # create rows
        rows = max( len(col1), len(col2), len(col3), len(inputdict) )

        if not rows:
            return

        if rows > len(col1):
            col1.extend(['']*(rows - len(col1)))
        if rows > len(col2):
            col2.extend(['']*(rows - len(col2)))
        if rows > len(col3):
            col3.extend(['']*(rows - len(col3)))

        keylist = list(inputdict.keys())
        if rows > len(keylist):
            keylist.extend([None]*(rows - len(keylist)))

        # keylist is a list of the dictionary keys, extended by None keys, if the dictionary is smaller than the number of rows of the table

        for index in range(rows):
            rownumber = index+header

            cssclass = rowc[index] if index < len(rowc) else ''
            if cssclass:
                self[rownumber] = tag.Part(tag_name='tr', attribs={"class":cssclass})
            else:
                self[rownumber] = tag.Part(tag_name='tr')

            if col1_class:
                self[rownumber][0] = tag.Part(tag_name='td', text = col1[index], attribs={"class":col1_class})
            else:
                self[rownumber][0] = tag.Part(tag_name='td', text = col1[index])

            if col2_class:
                self[rownumber][1] = tag.Part(tag_name='td', text = col2[index], attribs={"class":col2_class})
            else:
                self[rownumber][1] = tag.Part(tag_name='td', text = col2[index])

            if col3_class:
                self[rownumber][2] = tag.Part(tag_name='td', text = col3[index], attribs={"class":col3_class})
            else:
                self[rownumber][2] = tag.Part(tag_name='td', text = col3[index])

            if col4_class:
                self[rownumber][3] = tag.Part(tag_name='td', attribs={"class":col4_class})
            else:
                self[rownumber][3] = tag.Part(tag_name='td')

            if col4_style:
                self[rownumber][3].update_attribs({"style": col4_style})

            if up_style:
                up_button_style = up_style
            else:
                up_button_style = ''

            if up_hide:
                hide = up_hide[index] if index < len(up_hide) else False
                if hide:
                    if up_button_style:
                        up_button_style = up_button_style.strip(";")
                        up_button_style += ";visibility: hidden;"
                    else:
                        up_button_style = "visibility: hidden;"


            if up_button_style:
                self[rownumber][3][0] = tag.Part(tag_name="a", attribs={"role":"button", "style":up_button_style})
            else:
                self[rownumber][3][0] = tag.Part(tag_name="a", attribs={"role":"button"})

            if up_class:
                self[rownumber][3][0].update_attribs({"class": up_class})

            self[rownumber][3][0][0] = tag.HTMLSymbol("&uarr;")

            # create a url for the up arrow link
            upget1 = up_getfield1[index] if index < len(up_getfield1) else ''
            upget2 = up_getfield2[index] if index < len(up_getfield2) else ''
            get_fields = {self.get_formname("up_getfield1"):upget1,
                          self.get_formname("up_getfield2"):upget2}
            url = self.make_get_url(page, up_url, get_fields, True)
            self[rownumber][3][0].update_attribs({"href": url})

            if down_style:
                down_button_style = down_style
            else:
                down_button_style = ''

            if down_hide:
                hide = down_hide[index] if index < len(down_hide) else False
                if hide:
                    if down_button_style:
                        down_button_style = down_button_style.strip(";")
                        down_button_style += ";visibility: hidden;"
                    else:
                        down_button_style = "visibility: hidden;"

            if down_button_style:
                self[rownumber][3][1] = tag.Part(tag_name="a", attribs={"role":"button", "style":down_button_style})
            else:
                self[rownumber][3][1] = tag.Part(tag_name="a", attribs={"role":"button"})

            if down_class:
                self[rownumber][3][1].update_attribs({"class": down_class})

            self[rownumber][3][1][0] = tag.HTMLSymbol("&darr;")

            # create a url for the down arrow link
            downget1 = down_getfield1[index] if index < len(down_getfield1) else ''
            downget2 = down_getfield2[index] if index < len(down_getfield2) else ''
            get_fields = {self.get_formname("down_getfield1"):downget1,
                          self.get_formname("down_getfield2"):downget2}
            url = self.make_get_url(page, down_url, get_fields, True)
            self[rownumber][3][1].update_attribs({"href": url})

            key = keylist[index]
            if key:  # this is the dictionary key
                keyed_name = input_name + key
                self[rownumber][3][2] = tag.ClosedPart(tag_name="input", attribs={"name":keyed_name, "type":"hidden", "value":inputdict[key]})


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a click event handler"""
        jscript = """  $("#{ident} a").click(function (e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident = self.get_id())
        if self._up_jsonurl or self._down_jsonurl:
            return jscript + self._make_fieldvalues(upurl=self._up_jsonurl, downurl=self._down_jsonurl)



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<table>  <!-- with widget id and class widget_class -->
  <tr> <!-- with header class -->
    <th> <!-- title1 --> </th>
    <th> <!-- title2 --> </th>
    <th> <!-- title3 --> </th>
    <th> <!-- title4 --> </th>
  </tr>
  <tr>
    <td> <!-- with class col1_class and col1 text string --> </td>
    <td> <!-- with class col2_class and col2 text string --> </td>
    <td> <!-- with class col3_class and col3 text string --> </td>
    <td> <!-- with class col4_class -->
      <a role="button" href="#"> <!-- with style up_style -->
      <!-- The href link will be the url of the up ident with up get fields -->
      &uarr;
      </a>
      <a role="button" href="#"> <!-- with style down_style -->
      <!-- The href link will be the url of the down ident with down get fields -->
      &darr;
      </a>
      <input type="hidden" /> <!-- with name being 'widgetname:inputdict-keyname' and
                              <!-- value from inputdict values  -->
                              <!-- on form submission received data will be a widgfield -->
                              <!-- ('widgetname','inputdict') containing a dictionary of keyname:values -->
    </td>
  </tr>
  <!-- rows repeated -->
</table>"""


class InputTable3(Widget):
    """A table of four columns, two being text strings, one a text input field
       the fourth being a cell with up and down arrow links
       On the links being chosen a request for a json update is made, which
       typically updates the text columns, and the values in the input field.
       Each arrow has two get fields which can be set. A third getfield is attached to both
       arrows, and is automatically given the value set in the input field.
       The first row is four header titles.
       Typically inserted into a form, does not display errors"""

    display_errors = False


    arg_descriptions = {'header_class':FieldArg("cssclass",""),
                        'col1_class':FieldArg("cssclass",""),          # class applied to every td in the first column
                        'col2_class':FieldArg("cssclass",""),          # class applied to every td in the second column
                        'col3_class':FieldArg("cssclass",""),          # class applied to every td in the third column
                        'col4_class':FieldArg("cssclass",""),          # class applied to every td in the fourth column
                        'col4_style':FieldArg("cssstyle", ""),
                        'input_class':FieldArg("cssclass", ''),        # class applied to every input field in the third column
                        'input_style':FieldArg("cssstyle", ''),        # style applied to every input field in the third column
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'row_classes':FieldArgList('text', jsonset=True),
                        'title1':FieldArg('text', ''),
                        'title2':FieldArg('text', ''),
                        'title3':FieldArg('text', ''),
                        'title4':FieldArg('text', ''),
                        'up_link_ident':FieldArg("url", 'no_javascript'),
                        'up_json_ident':FieldArg("url", ''),
                        'down_link_ident':FieldArg("url", 'no_javascript'),
                        'down_json_ident':FieldArg("url", ''),
                        'up_style':FieldArg("cssstyle", ""),
                        'down_style':FieldArg("cssstyle", ""),
                        'up_class':FieldArg("cssclass", ""),
                        'down_class':FieldArg("cssclass", ""),
                        'up_hide':FieldArgList("boolean", jsonset=True),
                        'down_hide':FieldArgList("boolean", jsonset=True),
                        'up_getfield1':FieldArgList('text', valdt=True, jsonset=True),
                        'up_getfield2':FieldArgList('text', valdt=True, jsonset=True),
                        'down_getfield1':FieldArgList('text', valdt=True, jsonset=True),
                        'down_getfield2':FieldArgList('text', valdt=True, jsonset=True),
                        'getfield3':FieldArgList('text', valdt=True, jsonset=True),
                        'col1':FieldArgList('text', jsonset=True),
                        'col2':FieldArgList('text', jsonset=True),
                        'inputdict':FieldArgDict('text', valdt=True, jsonset=True, senddict=True)                 # dictionary of keyname:value
                        }                                                                                         # the field submitted as a dictionary
                                                                                                                  # each key will have 'keyname'
                                                                                                                  # and value will be the input field values

    def __init__(self, name=None, brief='', **field_args):
        """
        header_class: class of the header row, if empty string, then no class will be applied
        col1_class: class applied to every td in the first column
        col2_class: class applied to every td in the second column
        col3_class: class applied to every td in the third column
        col4_class: class applied to every td in the fourth column
        col4_style: style applied to every td in the fourth column
        input_class: class applied to every input field in the third column
        input_style: style applied to every input field in the third column
        size: The number of characters appearing in each text input area
        maxlength: The maximum number of characters accepted in each text area
        row_classes: A list of CSS classes to apply to each row (not including the header)
        title1: The header title over the first text column
        title2: The header title over the second text column
        title3: The header title over the third text column
        title4: The header title over the fourth text column
        up_link_ident: ident of the up arrow link if javascript disabled
        up_json_ident: ident of the up arrow link, expects a json file returned
        down_link_ident: ident of the down arrow link if javascript disabled
        down_json_ident: ident of the down arrow link, expects a json file returned
        up_style: CSS style applied to the up arrows
        down_style: CSS style applied to the down arrows
        up_class: CSS class applied to the up arrows
        down_class: CSS class applied to the down arrows
        up_hide: List of True/False to hide or not hide the up arrow
        down_hide: List of True/False to hide or not hide the down arrow
        up_getfield1: list of get fields, one for each up arrow link
        up_getfield2: list of second get fields, one for each up arrow link
        down_getfield1: list of get fields, one for each down arrow link
        down_getfield2: list of get fields, one for each down arrow link
        getfield3: list of values, which will be set on both arrows as getfield3 and will follow
                   the values set on the input field, Initially leave as an empty list, but
                   will be useful when doing a JSON update.
        col1: A list of text strings to place in the first column
        col2: A list of text strings to place in the second column
        inputdict: A dictionary of keyname:value, should be Ordered Dict unless python >= 3.6
                   the fields submitted will have name 'widgetname:inputdict-keyname'
                   and the user will receive a widgfield ('widgetname','inputdict') containing a dictionary of keyname:values
                   If values are None, then the values listed in getfield3 will be used, otherwise values given here have priority
         """
        Widget.__init__(self, name=name, tag_name="table", brief=brief, **field_args)
        self._up_jsonurl = ''
        self._down_jsonurl = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the table"
        inputdict = self.get_field_value("inputdict")
        rowc = self.get_field_value("row_classes")
        col1 = self.get_field_value("col1")
        col2 = self.get_field_value("col2")
        up_hide = self.get_field_value("up_hide")
        down_hide = self.get_field_value("down_hide")


        input_name = self.get_formname("inputdict") + '-'

        col1_class = self.get_field_value("col1_class")
        col2_class = self.get_field_value("col2_class")
        col3_class = self.get_field_value("col3_class")
        col4_class = self.get_field_value("col4_class")
        col4_style = self.get_field_value("col4_style")

        input_class = self.get_field_value("input_class")
        input_style = self.get_field_value("input_style")

        size = self.get_field_value('size')
        maxlength = self.get_field_value('maxlength')

        up_link_ident = self.get_field_value("up_link_ident") # ident of the up arrow link if javascript disabled
        if not up_link_ident:
            up_link_ident = 'no_javascript'
        up_json_ident = self.get_field_value("up_json_ident") # ident of the up arrow link, expects a json file returned

        down_link_ident = self.get_field_value("down_link_ident") # ident of the down arrow link if javascript disabled
        if not down_link_ident:
            down_link_ident = 'no_javascript'
        down_json_ident = self.get_field_value("down_json_ident") # ident of the down arrow link, expects a json file returned

        up_style = self.get_field_value("up_style") # CSS style applied to the up arrows
        down_style = self.get_field_value("down_style") # CSS style applied to the down arrows
        up_class = self.get_field_value("up_class") # CSS class applied to the up arrows
        down_class = self.get_field_value("down_class") # CSS class applied to the down arrows

        up_getfield1 = self.get_field_value("up_getfield1") # list of get fields, one for each up arrow link
        up_getfield2 = self.get_field_value("up_getfield2") # list of second get fields, one for each up arrow link
        down_getfield1 = self.get_field_value("down_getfield1") # list of get fields, one for each down arrow link
        down_getfield2 = self.get_field_value("down_getfield2") # list of get fields, one for each down arrow link

        getfield3 = self.get_field_value("getfield3") # list of get fields, one for each row, sent with both arrows

        # up arrow link
        # get json url
        if up_json_ident:
            self._up_jsonurl = skiboot.get_url(up_json_ident, proj_ident=page.proj_ident)
        # get url
        up_url = skiboot.get_url(up_link_ident, proj_ident=page.proj_ident)

        # down arrow link
        # get json url
        if down_json_ident:
            self._down_jsonurl = skiboot.get_url(down_json_ident, proj_ident=page.proj_ident)
        # get url
        down_url = skiboot.get_url(down_link_ident, proj_ident=page.proj_ident)


        header = 0
        if self.get_field_value('title1') or self.get_field_value('title2') or self.get_field_value('title3') or self.get_field_value('title4'):
            header = 1
            if self.get_field_value('header_class'):
                self[0] = tag.Part(tag_name='tr', attribs={"class":self.get_field_value('header_class')})
            else:
                self[0] = tag.Part(tag_name='tr')
            self[0][0] = tag.Part(tag_name='th', text = self.get_field_value('title1'))
            self[0][1] = tag.Part(tag_name='th', text = self.get_field_value('title2'))
            self[0][2] = tag.Part(tag_name='th', text = self.get_field_value('title3'))
            self[0][3] = tag.Part(tag_name='th', text = self.get_field_value('title4'))

        # create rows
        rows = max( len(col1), len(col2), len(inputdict) )

        if not rows:
            return

        if rows > len(col1):
            col1.extend(['']*(rows - len(col1)))
        if rows > len(col2):
            col2.extend(['']*(rows - len(col2)))

        keylist = list(inputdict.keys())
        if rows > len(keylist):
            keylist.extend([None]*(rows - len(keylist)))

        # keylist is a list of the dictionary keys, extended by None keys, if the dictionary is smaller than the number of rows of the table

        for index in range(rows):
            rownumber = index+header

            cssclass = rowc[index] if index < len(rowc) else ''
            if cssclass:
                self[rownumber] = tag.Part(tag_name='tr', attribs={"class":cssclass})
            else:
                self[rownumber] = tag.Part(tag_name='tr')

            if col1_class:
                self[rownumber][0] = tag.Part(tag_name='td', text = col1[index], attribs={"class":col1_class})
            else:
                self[rownumber][0] = tag.Part(tag_name='td', text = col1[index])

            if col2_class:
                self[rownumber][1] = tag.Part(tag_name='td', text = col2[index], attribs={"class":col2_class})
            else:
                self[rownumber][1] = tag.Part(tag_name='td', text = col2[index])

            if col3_class:
                self[rownumber][2] = tag.Part(tag_name='td', attribs={"class":col3_class})
            else:
                self[rownumber][2] = tag.Part(tag_name='td')

            # getfield3 is used for both arrows, and also for the input fields
            get3 = getfield3[index] if index < len(getfield3) else ''

            key = keylist[index]
            if key:  # this is the dictionary key
                keyed_name = input_name + key
                if not inputdict[key]:
                    # if no value given, and if get3 has a value, that is used
                    if get3:
                        inputdict[key] = get3
                else:
                    # inputdict overrides get3, but the two must match
                    get3 = inputdict[key]

                # set up the input field, including an onchange event which sets getfield3 when the input field changes
                self[rownumber][2][0] = tag.ClosedPart(tag_name="input", attribs={"name":keyed_name,
                                                                                  "type":"text",
                                                                                   "value":inputdict[key],
                "onchange":"SKIPOLE.widgets['{ident}'].setnewnumber(this.value, {data})".format(ident = self.get_id(), data=rownumber)
                                                                                 })
                if size:
                    self[rownumber][2][0].update_attribs({"size":size})
                if maxlength:
                    self[rownumber][2][0].update_attribs({"maxlength":maxlength})
                if input_class:
                    self[rownumber][2][0].update_attribs({"class":input_class})
                if input_style:
                    self[rownumber][2][0].update_attribs({"style":input_style})

            if col4_class:
                self[rownumber][3] = tag.Part(tag_name='td', attribs={"class":col4_class})
            else:
                self[rownumber][3] = tag.Part(tag_name='td')

            if col4_style:
                self[rownumber][3].update_attribs({"style": col4_style})

            if up_style:
                up_button_style = up_style
            else:
                up_button_style = ''

            if up_hide:
                hide = up_hide[index] if index < len(up_hide) else False
                if hide:
                    if up_button_style:
                        up_button_style = up_button_style.strip(";")
                        up_button_style += ";visibility: hidden;"
                    else:
                        up_button_style = "visibility: hidden;"

            if up_button_style:
                self[rownumber][3][0] = tag.Part(tag_name="a", attribs={"role":"button", "style":up_button_style})
            else:
                self[rownumber][3][0] = tag.Part(tag_name="a", attribs={"role":"button"})

            if up_class:
                self[rownumber][3][0].update_attribs({"class": up_class})

            self[rownumber][3][0][0] = tag.HTMLSymbol("&uarr;")

            # create a url for the up arrow link
            upget1 = up_getfield1[index] if index < len(up_getfield1) else ''
            upget2 = up_getfield2[index] if index < len(up_getfield2) else ''
            get_fields = {self.get_formname("up_getfield1"):upget1,
                          self.get_formname("up_getfield2"):upget2,
                          self.get_formname("getfield3"):get3
                          }
            url = self.make_get_url(page, up_url, get_fields, True)
            self[rownumber][3][0].update_attribs({"href": url})

            if down_style:
                down_button_style = down_style
            else:
                down_button_style = ''

            if down_hide:
                hide = down_hide[index] if index < len(down_hide) else False
                if hide:
                    if down_button_style:
                        down_button_style = down_button_style.strip(";")
                        down_button_style += ";visibility: hidden;"
                    else:
                        down_button_style = "visibility: hidden;"

            if down_button_style:
                self[rownumber][3][1] = tag.Part(tag_name="a", attribs={"role":"button", "style":down_button_style})
            else:
                self[rownumber][3][1] = tag.Part(tag_name="a", attribs={"role":"button"})

            if down_class:
                self[rownumber][3][1].update_attribs({"class": down_class})

            self[rownumber][3][1][0] = tag.HTMLSymbol("&darr;")

            # create a url for the down arrow link
            downget1 = down_getfield1[index] if index < len(down_getfield1) else ''
            downget2 = down_getfield2[index] if index < len(down_getfield2) else ''
            get_fields = {self.get_formname("down_getfield1"):downget1,
                          self.get_formname("down_getfield2"):downget2,
                          self.get_formname("getfield3"):get3
                          }
            url = self.make_get_url(page, down_url, get_fields, True)
            self[rownumber][3][1].update_attribs({"href": url})




    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a click event handler"""
        jscript = """  $("#{ident} a").click(function (e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident = self.get_id())
        if self._up_jsonurl or self._down_jsonurl:
            return jscript + self._make_fieldvalues(upurl=self._up_jsonurl, downurl=self._down_jsonurl)



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<table>  <!-- with widget id and class widget_class -->
  <tr> <!-- with header class -->
    <th> <!-- title1 --> </th>
    <th> <!-- title2 --> </th>
    <th> <!-- title3 --> </th>
    <th> <!-- title4 --> </th>
  </tr>
  <tr>
    <td> <!-- with class col1_class and col1 text string --> </td>
    <td> <!-- with class col2_class and col2 text string --> </td>
    <td> <!-- with class col3_class -->
      <input type="text" />   <!-- class input_class, style input_style -->
                              <!-- with name being 'widgetname:inputdict-keyname' and -->
                              <!-- value from inputdict values  -->
                              <!-- on form submission received data will be a widgfield -->
                              <!-- ('widgetname','inputdict') containing a dictionary of keyname:values -->
    </td>
    <td> <!-- with class col4_class -->
      <a role="button" href="#"> <!-- with style up_style -->
      <!-- The href link will be the url of the up ident with up get fields -->
      <!-- and also getfield3 which reflect the value shown on the input field -->
      &uarr;
      </a>
      <a role="button" href="#"> <!-- with style down_style -->
      <!-- The href link will be the url of the down ident with down get fields -->
      <!-- and also getfield3 which reflect the value shown on the input field -->
      &darr;
      </a>
    </td>
  </tr>
  <!-- rows repeated -->
</table>"""








import html

from collections import OrderedDict

from skipole import ValidateError, FailPage, ServerError, GoTo
from skipole import skilift
from skipole.skilift import editpage, editfolder, fromjson
from .. import utils


def edit_root(skicall):
    "Go to edit root folder, with no other call_data contents other than those set here"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # called by responder 3, edit_root link from Root Folder nav button
    # this responder then targets responder 22008 to fill in root folder fields
    utils.clear_call_data(call_data)
    call_data['folder_number'] = 0


def goto_edited_folder(skicall):
    """Called by responder 2001 with requests from various places to to edit a folder
       This responder passes the call to another responder which calls retrieve_edited_folder"""

    call_data = skicall.call_data
    page_data = skicall.page_data
    submit_dict = skicall.submit_dict

    if 'received_data' not in submit_dict:
        raise FailPage(message = "Folder missing")

    # from the folder edit page
    widgfields = submit_dict['received_data']
    if ('ftree','edited_item') in widgfields:
        call_data['edit_folder'] = widgfields['ftree','edited_item']


def retrieve_edited_folder(skicall):
    "Fills in the edit folder page, including the tree of folder contents"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']

    # 'edit_folder' is from a form, not from session data

    if 'edit_folder' in call_data:
        try:
            proj,folder = call_data['edit_folder'].split('_')
            folder_number = int(folder)
        except Exception:
            raise FailPage(message = "Invalid Folder")
        if proj != editedprojname:
            raise FailPage(message = "Invalid Folder")
        del call_data['edit_folder']
        call_data['folder_number'] = folder_number
    elif 'folder_number' in call_data:
        folder_number = call_data['folder_number']
    else:
        raise FailPage(message = "Folder missing")

    info = skilift.item_info(editedprojname, folder_number)

    # info is a named tuple with members
    # 'project', 'project_version', 'itemnumber', 'item_type', 'name', 'brief', 'path', 'label_list', 'change', 'parentfolder_number', 'restricted'

    if (not info) or (info.item_type != "Folder"):
        raise FailPage(message = "Invalid folder")

    call_data['fchange'] = info.change

    contents, dragrows, droprows = _foldertree(editedprojname, folder_number)
    page_data['ftree', 'contents'] = contents

    # old code, was just rows 

    page_data['ftree', 'dragrows'] = dragrows
    page_data['ftree', 'droprows'] = droprows
    page_data['ftree', 'dropident'] = 'drop_rows' # label of responder

    # cols
    # A list of two element lists for every column in the table:
    # 0 The target HTML ident or label of buttons in the column, if the second element of this list is empty string or the client has javascript disabled.
    # 1 The target JSON ident or label of buttons in the column.

    page_data['ftree', 'cols'] = [ ["", ""],
                                   ["", ""],
                                   ["", ""],
                                   ["", ""],
                                   ["edit_action", ""],
                                   ["edit_action", ""],
                                   ["edit_action", ""],
                                   ["no_javascript", "2003"],           # calls responder which calls copy_page
                                   ["no_javascript", "2004"]  ]


    # set edited folder in the required dictionary
    page_data['ie1:page_ident'] = (info.project, info.itemnumber)
    if 'brief' in call_data:
        page_data['st2:folder_brief'] = call_data['brief']
    else:
        page_data['st2:folder_brief'] = info.brief

    # check if this is rootfolder
    if info.itemnumber == 0:
        page_data[("adminhead","page_head","large_text")] = "Edit Root Folder : %s" % (info.path,)
        # hide certain items if this is the root folder
        page_data['rename_folder','show'] = False
        page_data['sp1:show_parent_restricted'] = False
        page_data['sb1:show_set_restricted'] = False
        page_data['sb2:show_set_unrestricted'] = False
        page_data['aboutdeletefolder','show'] = False
        page_data['deletefolder','show'] = False
    else:
        page_data[("adminhead","page_head","large_text")] = "Edit Folder : %s" % (info.path,)
        # The folder can be given a name
        if 'name' in call_data:
            page_data['rename_folder','input_text'] = call_data['name']
        else:
            page_data['rename_folder', 'input_text'] = info.name
        parent = skilift.item_info(info.project, info.parentfolder_number)
        if parent.restricted:
            page_data['sp1:show_parent_restricted'] = True
            page_data['sb1:show_set_restricted'] = False
            page_data['sb2:show_set_unrestricted'] = False
        else:
            # parent is not restricted
            page_data['sp1:show_parent_restricted'] = False
            if info.restricted:
                # folder currently restricted
                page_data['sb1:show_set_restricted'] = False
                page_data['sb2:show_set_unrestricted'] = True
            else:
                # folder currently unrestricted
                page_data['sb1:show_set_restricted'] = True
                page_data['sb2:show_set_unrestricted'] = False
    # get default page, and list of page names within the folder
    default_page_name, page_names = skilift.folder_page_names(info.project, info.itemnumber)
    if default_page_name:
        page_data['sdd1:selectvalue'] = default_page_name
    else:
        page_data['sdd1:selectvalue'] = "-None-"
    page_data['sdd1:option_list'] = page_names
    if page_names:
        if info.restricted:
            # do not give a choose default page option if the folder has restricted access
            page_data['sdd1:show'] = False
        else:
            page_data['sdd1:option_list'].insert(0,"-None-")
            page_data['sdd1:show'] = True
    else:
        # Only give a default page option if pages are present
        page_data['sdd1:show'] = False


def move_to_folder(skicall):
    "moves item defined by dragrows to folder defined by droprows and refreshes ftree via JSON call"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']

    # folder_number is the top folder of ftree
    if 'folder_number' in call_data:
        folder_number = call_data['folder_number']
    else:
        raise FailPage(message = "Folder missing")

    # item_ident is the item to move
    item_ident = call_data['ftree', 'dragrows']

    if '_' not in item_ident:
        raise FailPage(message="No valid item to move given")

    item_proj, item_num = item_ident.split("_")
    if item_proj != editedprojname:
        raise FailPage(message="No valid item to move given")

    # folder_ident is the target folder
    folder_ident = call_data['ftree', 'droprows']

    if '_' not in folder_ident:
        raise FailPage(message="No valid target folder given")

    folder_proj, folder_num = folder_ident.split("_")
    if folder_proj != editedprojname:
        raise FailPage(message="No valid target folder given")

    try:
        folder_num = int(folder_num)
        item_num = int(item_num)
    except Exception:
        raise FailPage(message="Submission not recognised - integer numbers of item idents are require")

    # call function in skilift.editfolder
    try:
        editfolder.move_to_folder(editedprojname, item_num, folder_num)
    except ServerError as e:
        raise FailPage(message=e.message)
    call_data['status'] = 'Item moved'
    # refresh ftree
    contents, dragrows, droprows = _foldertree(editedprojname, folder_number)
    page_data['ftree', 'contents'] = contents
    page_data['ftree', 'dragrows'] = dragrows
    page_data['ftree', 'droprows'] = droprows
    # this is sent by json to the page


def choose_edit_action(skicall):
    "Choose which action to take when called from ftree"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('ftree', 'contents') not in call_data:
        raise FailPage(message = "Requested action not recognised")
    action = call_data['ftree', 'contents']
    try:
        get_field_list = action.split('_')
        project = get_field_list[-2]
        itemnumber = int(get_field_list[-1])
    except Exception:
        raise FailPage(message = "Requested action not recognised")
    if action.startswith('edit_folder_'):
        call_data['edit_folder'] = action[12:]
        raise GoTo(('skiadmin',22008), clear_submitted=True, clear_page_data=True)
    elif action.startswith('add_folder_'):
        call_data['parent'] = action[11:]
        raise GoTo(('skiadmin',22107), clear_submitted=True, clear_page_data=True)
    elif action.startswith('add_page_'):
        call_data['edited_folder'] = action[9:]
        raise GoTo(('skiadmin',22307), clear_submitted=True, clear_page_data=True)
    elif action.startswith('edit_page_'):
        # must find type of page to edit
        try:
            # pinfo is tuple 'name', 'number', 'restricted', 'brief', 'item_type', 'responder', 'enable_cache', 'change'
            pinfo = skilift.page_info(project, itemnumber)
        except Exception:
            raise FailPage(message = "Requested page not recognised")
        call_data['page_number'] = pinfo.number
        call_data['pchange'] = pinfo.change
        if pinfo.item_type == 'CSS':
            raise GoTo(28007, clear_submitted=True, clear_page_data=True)         # css edit page
        elif pinfo.item_type == 'JSON':
            raise GoTo(20407, clear_submitted=True, clear_page_data=True)         # json edit page
        elif pinfo.item_type == 'TemplatePage':
            raise GoTo(23207, clear_submitted=True, clear_page_data=True)          # retrieve template edit page contents
        elif pinfo.item_type == 'SVG':
            raise GoTo(23407, clear_submitted=True, clear_page_data=True)          # svg edit page
        elif pinfo.item_type == 'RespondPage':
            raise GoTo(26007, clear_submitted=True, clear_page_data=True)          # retrieve responder edit page contents
        elif pinfo.item_type == 'FilePage':
            raise GoTo(29007, clear_submitted=True, clear_page_data=True)          # filepage edit page
    # no action recognised
    raise FailPage(message = "Requested action not recognised")



def choose_remove_action(skicall):
    "Choose to remove a page or folder, fills in confirm box"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('ftree', 'contents') not in call_data:
        raise FailPage(message = "Requested action not recognised")
    action = call_data['ftree', 'contents']
    try:
        get_field_list = action.split('_')
        project = get_field_list[-2]
        itemnumber = int(get_field_list[-1])
        info = skilift.item_info(project, itemnumber)
    except Exception:
        raise FailPage(message = "Item to remove not recognised")
    if not info:
        raise FailPage(message = "Item to remove not recognised")
    if itemnumber == 0:
        raise FailPage(message = "Cannot delete the root folder")
    # display the modal confirm box
    page_data['page_delete', 'hide'] = False
    page_data['page_delete', 'para_text'] = "Delete %s with ident %s and name %s" % (info.item_type, info.itemnumber, info.name)
    page_data['page_delete', 'get_field2_1'] = project + '_' + str(itemnumber)


def copy_page(skicall):
    """If chosen from a page row, copies a page, stores into localStorage
       If chosen from a folder row, pastes the copied page"""
    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('ftree', 'contents') not in call_data:
        raise FailPage(message = "Requested action not recognised")
    action = call_data['ftree', 'contents']
    try:
        get_field_list = action.split('_')
        project = get_field_list[-2]
        itemnumber = int(get_field_list[-1])
        info = skilift.item_info(project, itemnumber)
    except Exception:
        raise FailPage(message = "Item to copy not recognised")
    if not info:
        raise FailPage(message = "Item to copy not recognised")
    if info.item_type != "Folder":
        # Its a page
        try:
            jsonstring = fromjson.page_to_json(project, itemnumber, template_svg_only=False)
        except ServerError as e:
            raise FailPage(message=e.message)
        page_data['localStorage'] = {'ski_page':jsonstring}
        call_data['status'] = "Page %s is copied, choose Paste Page on a Folder." % itemnumber
        return
    # So item_type is Folder, the paste button has been pressed, display modal box requesting a new id number
    page_data['pastediv', 'hide'] = False
    # set next available ident number in pastepage
    page_data['pastepage', 'input_text'] = str(skilift.next_ident_number(project))
    page_data['pastepage', 'hidden_field1'] = action


def paste_page(skicall):
    "Called to paste a page into the folder"
    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']

    if ('pastepage', 'hidden_field1') not in call_data:
        raise FailPage(message = "Requested action not recognised", widget='pastepage')
    action = call_data['pastepage', 'hidden_field1']
    try:
        get_field_list = action.split('_')
        project = get_field_list[-2]
        if project != editedprojname:
            raise FailPage(message = "Folder not recognised", widget='pastepage')
        itemnumber = int(get_field_list[-1])
        info = skilift.item_info(project, itemnumber)
    except Exception:
        raise FailPage(message = "Folder not recognised", widget='pastepage')
    if not info:
        raise FailPage(message = "Folder not recognised", widget='pastepage')
    if info.item_type != "Folder":
        raise FailPage(message = "Can only paste into a Folder!", widget='pastepage')
    if ('pastepage', 'input_text') not in call_data:
        pagenumber = skilift.next_ident_number(project)
    else:
        try:
            pagenumber = int(call_data['pastepage', 'input_text'])
        except:
            raise FailPage(message = "New page number must be an integer", widget='pastepage')
    # check pagenumber is valid
    if pagenumber < 1:
        raise FailPage(message = "The page number is invalid", widget='pastepage')
    # check if an item with this page number already exists
    item = skilift.item_info(project, pagenumber)
    if item is not None:
        raise FailPage(message = "This page number is already used (%s)" % (item.path,), widget='pastepage')
    # get pages in the folder
    default_page_name, page_names = skilift.folder_page_names(project, itemnumber)
    base_page_name = "newpage"
    new_page_name = "newpage"
    new_page_count = 1
    while True:
        if new_page_name in page_names:
            new_page_count += 1
            new_page_name = base_page_name + str(new_page_count)
        else:
            break
    call_data['NewName'] = new_page_name
    call_data['NewBrief'] = "New page %s" % pagenumber
    if ('pastepage', 'local_storage') not in call_data:
        # divert to page to input ident number of page to be copied - 22461
        call_data['InFolder'] = itemnumber
        call_data['NewPage'] = pagenumber
        _retrieve_new_copypage(skicall)
        raise GoTo(target=(call_data['adminproj'], 22461))
    else:
        # we have the page data
        # Get the page to be copied
        json_string = call_data['pastepage', 'local_storage']
        # create the page
        try:
            fromjson.create_page(project, itemnumber, pagenumber, new_page_name, call_data['NewBrief'], json_string)
            fchange = editfolder.folderchange(project, itemnumber)
        except ServerError as e:
            raise FailPage(message = e.message, widget='pastepage')
        # folder_number and fchange are held in session data, if these have been changed, set into session data
        if "folder_number" in call_data:
            folder_number = call_data["folder_number"]
            if folder_number == itemnumber:
                call_data["fchange"] = fchange
        else:
            folder_number = itemnumber
            call_data["folder_number"] = folder_number
            call_data["fchange"] = fchange

        # This data sent in general_json 
        contents, dragrows, droprows = _foldertree(editedprojname, folder_number)
        page_data['ftree', 'contents'] = contents
        page_data['ftree', 'dragrows'] = dragrows
        page_data['ftree', 'droprows'] = droprows
        # hide the pastediv and show status
        call_data['status'] = 'Page %s added' % (new_page_name,)
        page_data['pastediv', 'hide'] = True
        page_data['ClearAllErrors'] = True


def _retrieve_new_copypage(skicall):
    """Gets data for a create a page copy - variation of similar function in addpage.py
       Fills in the details for the page requesting the ident, or upload of a page to be copied"""

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']

    new_name = call_data['NewName']
    new_brief = call_data['NewBrief']
    
    parent_url = skilift.page_path(editedprojname, call_data['InFolder'])

    page_data[("adminhead","page_head","large_text")] = "Add a page copy to : %s" % (parent_url,)

    # information paragraphs
    page_data['page_name_text:para_text'] = "New page name : " + new_name
    page_data['page_brief_text:para_text'] = "Description   : " + new_brief

    # information hidden fields
    page_data['copyident','pagename'] = new_name
    page_data['copyident','pagebrief'] = new_brief
    page_data['copyident','pageident'] = str(call_data['NewPage'])
    page_data['copyident','parent'] = str(call_data['InFolder'])

    page_data['upload','pagename'] = new_name
    page_data['upload','pagebrief'] = new_brief
    page_data['upload','pageident'] = str(call_data['NewPage'])
    page_data['upload','parent'] = str(call_data['InFolder'])

    # top descriptive text
    page_data['top_text:para_text'] = """To copy a page, either the ident number of an existing page to be copied is required, alternatively - a previously downloaded page definition file can be uploaded."""

    # the submit ident text input
    if 'copyident' in call_data:
        page_data['copyident','input_text'] = call_data['copyident']


def delete_item(skicall):
    "Deletes a folder or page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    try:
        ident = call_data['page_delete','get_field2_1'].split('_')
        project = ident[0]
        itemnumber = int(ident[1])
        info = skilift.item_info(project, itemnumber)
    except Exception:
        raise FailPage(message = "Item to remove not recognised")

    if not info:
        raise FailPage(message = "Item to remove not recognised")

    parentfolder_number = info.parentfolder_number

    if itemnumber == 0:
        raise FailPage(message = "Cannot delete the root folder")

    if info.item_type == 'Folder':
        # call skilift.editfolder.delete_folder
        error = editfolder.delete_folder(project, itemnumber)
        if error:
            raise FailPage(message = "Unable to delete the folder")
        call_data['status'] = 'Folder deleted'
        # get folder to show on ftree
        if "folder_number" in call_data:
            folder_number = call_data["folder_number"]
            if folder_number == itemnumber:
                folder_number = parentfolder_number
                call_data["folder_number"] = parentfolder_number
        else:
            folder_number = parentfolder_number
            call_data["folder_number"] = parentfolder_number
        call_data["fchange"] = editfolder.folderchange(project, folder_number)
        return

    # If not a Folder, must be a page

    # call skilift.editpage.delete_page
    error = editpage.delete_page(project, itemnumber)
    if error:
        raise FailPage(message = "Unable to delete the page")

    # as page is deleted, this will affect any page objects
    # stored in call_data, so clear them

    if 'page_number' in call_data:
        del call_data['page_number']
    if 'page' in call_data:
        del call_data['page']

    # get folder to show on ftree
    if "folder_number" in call_data:
        folder_number = call_data["folder_number"]
    else:
        folder_number = parentfolder_number
        call_data["folder_number"] = parentfolder_number
    call_data["fchange"] = editfolder.folderchange(project, folder_number)
    call_data['status'] = 'Page deleted'


def submit_rename_folder(skicall):
    "rename this folder"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if 'folder_number' not in call_data:
        raise FailPage(message = "Folder missing")
    if 'name' not in call_data:
        raise FailPage(message="No new folder name has been given", widget="rename_folder")
    # Rename the folder
    # call skilift.editfolder.rename_folder which returns a new fchange
    try:
        call_data['fchange'] = editfolder.rename_folder(call_data['editedprojname'], call_data['folder_number'], call_data['fchange'], call_data['name'])
    except ServerError as e:
        raise FailPage(e.message, widget="rename_folder")
    call_data['status'] = 'Folder renamed'


def submit_folder_brief(skicall):
    "set this folders brief"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if 'folder_number' not in call_data:
        raise FailPage(message = "Folder missing")
    project = call_data['editedprojname']
    foldernumber = call_data['folder_number']
    fchange = call_data['fchange']
    if 'brief' not in call_data:
        raise FailPage(message="No folder description available", widget="st2")
    new_brief = call_data['brief']
    if not new_brief:
        raise FailPage(message="No folder brief given", widget="st2")
    # call skilift.editfolder.folder_description which returns a new fchange
    try:
        call_data['fchange'] = editfolder.folder_description(project, foldernumber, fchange, new_brief)
    except ServerError as e:
        raise FailPage(e.message, widget="st2")
    call_data['status'] = 'Folder description set : %s' % (new_brief,)


def submit_default_page(skicall):
    "Set this folder's default page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if 'folder_number' not in call_data:
        raise FailPage(message = "Folder missing")
    project = call_data['editedprojname']
    foldernumber = call_data['folder_number']
    fchange = call_data['fchange']
    if 'selectvalue' not in call_data:
        raise FailPage(message="The page to set as default has not been found", widget="sdd1")
    if call_data['selectvalue'] == "-None-":
        default_page_name = ""
    else:
        default_page_name = call_data['selectvalue']
    try:
        call_data['fchange'] = editfolder.set_default_page(project, foldernumber, fchange, default_page_name)
    except ServerError as e:
        raise FailPage(e.message, widget="sdd1")
    call_data['status'] = 'Default page set'


def submit_restricted(skicall):
    "set this folder as restricted"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if 'folder_number' not in call_data:
        raise FailPage(message = "Folder missing")
    # restrict the folder
    try:
        call_data['fchange'] = editfolder.set_restricted_status(call_data['editedprojname'], call_data['folder_number'], call_data['fchange'], True)
    except ServerError as e:
        raise FailPage(e.message)


def submit_unrestricted(skicall):
    "set this folder as unrestricted"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if 'folder_number' not in call_data:
        raise FailPage(message = "Folder missing")
    # un-restrict the folder
    try:
        call_data['fchange'] = editfolder.set_restricted_status(call_data['editedprojname'], call_data['folder_number'], call_data['fchange'], False)
    except ServerError as e:
        raise FailPage(e.message)


def downloadfolder(skicall):
    "Gets folder, and returns a json dictionary, this will be sent as an octet file to be downloaded"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if 'folder_number' not in call_data:
        raise FailPage(message = "Folder missing")
    jsonstring =  fromjson.folder_to_json(call_data['editedprojname'], call_data['folder_number'], indent=4)
    line_list = []
    n = 0
    for line in jsonstring.splitlines(True):
        binline = line.encode('utf-8')
        n += len(binline)
        line_list.append(binline)
    page_data['headers'] = [('content-type', 'application/octet-stream'), ('content-length', str(n))]
    return line_list


def confirmrecursive(skicall):
    "Displays the confirm recursive modal box"
    skicall.page_data['recursive_delete', 'hide'] = False


def recursivedelete(skicall):
    "Delete the folder and contents"
    call_data = skicall.call_data
    page_data = skicall.page_data

    if 'folder_number' not in call_data:
        raise FailPage(message = "Folder missing")

    folder_number = call_data['folder_number']
    editedprojname = call_data['editedprojname']
    fchange = call_data['fchange']

    folderinfo = skilift.folder_info(editedprojname, folder_number)

    if fchange != folderinfo.change:
        call_data['fchange'] = folderinfo.change
        raise FailPage("The folder has been changed prior to this submission, someone else may be editing this project")

    # As the current folder is to be deleted, clear out session data from call_data
    utils.clear_call_data(call_data)

    # Delete the folder, get num and change of parentfolder
    try:
        num, change = editfolder.delete_folder_recursively(editedprojname, folder_number, fchange)
    except ServerError as e:
        raise FailPage(e.message)

    # replace folder_number and fchange in call_data for session values
    call_data['folder_number'] = num
    call_data['fchange'] = change
    call_data['status'] = 'Folder %s, number:%s and contents deleted' % (folderinfo.name, folder_number)




def _foldertree(projectname, foldernumber):
    """Returns contents list of lists with ftree table fields, with the given foldernumber of the folder at the top of
    the table"""

    finfo = skilift.folder_info(projectname, foldernumber)
    folder_path = skilift.item_info(projectname, foldernumber).path
    folder_ident = projectname + '_' + str(foldernumber)

    #   col 0 - text string, either text to display or button text
    #   col 1 - A 'style' string set on the td cell, if empty string, no style applied
    #   col 2 - Is button? If False only text will be shown, not a button, button class will not be applied
    #           If True a link to link_ident/json_ident will be set with button_class applied to it
    #   col 3 - The get field value of the button link, empty string if no get field

    # This creates a contents of cell's, each row of the table has eight columns
    contents = []
    dragrows = []
    droprows = []

    # first cell is the folder URL, no style, Not a link, no get field
    contents.append( (folder_path, '', False, '') )
 
    # second cell is the folder ident number
    contents.append( (str(finfo.number), '', False, '') )

    # third cell is restricted or not
    if finfo.restricted:
        contents.append( ('R', 'width : 1%;text-align: center;color: black;background-color: red;', False, '') )
    else:
        contents.append( ('', '', False, '') )

    # fourth cell is the folder brief
    if len(finfo.brief)>40:
        contents.append( (html.escape(finfo.brief[:35] + "..."), '', False, '') )
    else:
        contents.append( (html.escape(finfo.brief[:35]), '', False, '') )

    # fifth cell is an Edit link
    # top folder does not have an edit link
    contents.append( ('', '', False, '') )

    # sixth cell is add folder
    contents.append( ('Add Folder', 'width : 1%;text-align: center;', True, 'add_folder_' + folder_ident)    )

    # seventh cell is add a new page
    contents.append( ('New Page', 'width : 1%;text-align: center;', True, 'add_page_' + folder_ident) )

    # eighth cell is paste page
    contents.append( ('Paste Page', 'width : 1%;text-align: center;', True, 'paste_page_' + folder_ident) )

    # ninth cell is remove line - but no remove link for the top line
    contents.append( ('', '', False, '') )

    dragrows.append([False, ""])
    droprows.append([True, folder_ident])

    # place all sub pages in rows beneath the folder
    if finfo.contains_pages:
        _show_pages(contents, projectname, foldernumber, dragrows, droprows, 2)
    if finfo.contains_folders:
        _show_folders(contents, projectname, foldernumber, dragrows, droprows, 2)
    return contents, dragrows, droprows


def _show_pages(contents, projectname, foldernumber, dragrows, droprows, indent):
    """Used to create pages  beneath the folder"""

    # pinfo attributes are 'name', 'number', 'restricted', 'brief', 'item_type', 'responder'

    ident = projectname + "_"
    padding = "padding-left : %sem;" % (indent,)

    for pinfo in skilift.pages(projectname, foldernumber):

        page_ident = ident + str(pinfo.number)
        dragrows.append([True, page_ident])
        droprows.append([False, ""])

        # first column is the page name, style includes padding, not a link, no get field
        contents.append( (pinfo.name, padding, False, '') )
 
        # second column is the page ident number, no style, not a link, no get field
        contents.append( (str(pinfo.number), '', False, '') )

        # third column is restricted or not
        if pinfo.restricted:
            contents.append( ('R', 'width : 1%;text-align: center;color: black;background-color: red;', False, '') )
        else:
            contents.append( ('', '', False, '') )

        # fourth cell is the page brief
        if len(pinfo.brief)>40:
            contents.append( (html.escape(pinfo.brief[:35] + "..."), '', False, '') )
        else:
            contents.append( (html.escape(pinfo.brief[:35]), '', False, '') )

        # fifth column is an Edit link
        contents.append( ('Edit', 'width: 1%;text-align: center;', True, 'edit_page_' + page_ident) )

        # sixth column is page type
        # seventh either empty or responder type

        if pinfo.item_type == 'TemplatePage':
            contents.append( ('Template', 'text-align: center;', False, '') )
            contents.append( ('', '', False, '') )
        elif pinfo.item_type == 'RespondPage':
            contents.append( ('Responder', 'text-align: center;', False, '') )
            contents.append( (pinfo.responder, 'text-align: center;', False, '') )
        elif pinfo.item_type == 'CSS':
            contents.append( ('CSS', 'text-align: center;', False, '') )
            contents.append( ('', '', False, '') )
        elif pinfo.item_type == 'JSON':
            contents.append( ('JSON', 'text-align: center;', False, '') )
            contents.append( ('', '', False, '') )
        elif pinfo.item_type == 'SVG':
            contents.append( ('SVG', 'text-align: center;', False, '') )
            contents.append( ('', '', False, '') )
        elif pinfo.item_type == 'FilePage':
            contents.append( ('File', 'text-align: center;', False, '') )
            contents.append( ('', '', False, '') )
        else:
            raise ServerError(message="An unknown page type")

        # eighth column is copy page
        contents.append( ('Copy', 'width : 1%;text-align: center;', True, page_ident) )

        # ninth column is remove page
        contents.append( ('Remove', 'width : 1%;text-align: center;', True, page_ident) )

    return


def _show_folders(contents, projectname, foldernumber, dragrows, droprows, indent):


    # finfo attributes are 'name', 'number', 'restricted', 'brief', 'contains_pages', 'contains_folders'

    ident = projectname + "_"
    padding = "padding-left : %sem;" % (indent,)

    for finfo in skilift.folders(projectname, foldernumber):

        folder_ident = ident + str(finfo.number)
        dragrows.append([True, folder_ident])
        droprows.append([True, folder_ident])

        # first column is the folder path from parent, with padding style, Not a link, no get field
        contents.append( (finfo.name+"/", padding, False, '') )

        # second column is the folder ident number, no style, not a link, no get field
        contents.append( (str(finfo.number), '', False, '') )

        # third column is restricted or not
        if finfo.restricted:
            contents.append( ('R', 'width : 1%;text-align: center;color: black;background-color: red;', False, '') )
        else:
            contents.append( ('', '', False, '') )

        # fourth cell is the folder brief
        if len(finfo.brief)>40:
            contents.append( (html.escape(finfo.brief[:35] + "..."), '', False, '') )
        else:
            contents.append( (html.escape(finfo.brief[:35]), '', False, '') )

        # fifth column is an Edit link
        contents.append( ('Edit', 'width: 1%;text-align: center;', True, 'edit_folder_' + folder_ident) )

        # sixth column is add folder
        contents.append( ('Add Folder', 'width : 1%;text-align: center;', True, 'add_folder_' + folder_ident) )

        # seventh column is add a new page
        contents.append( ('New Page', 'width : 1%;text-align: center;', True, 'add_page_' + folder_ident) )

        # eighth cell is paste page
        contents.append( ('Paste Page', 'width : 1%;text-align: center;', True, 'paste_page_' + folder_ident) )

        # ninth column is remove line - but no remove if folder has contents
        if finfo.contains_pages or finfo.contains_folders:
            contents.append( ('', '', False, '') )
        else:
            contents.append( ('Remove', 'width : 1%;text-align: center;', True, folder_ident) )

        # place all sub pages in rows beneath the subfolder
        if finfo.contains_pages:
            _show_pages(contents, projectname, finfo.number, dragrows, droprows, indent+1)

        if finfo.contains_folders:
            _show_folders(contents, projectname, finfo.number, dragrows, droprows, indent+1)

    return








"""Contains widgets for logging in to a service"""


from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class Pin4(Widget):
    """Defines a form containing four single character input fields."""


    _container = ((0,1),)

    error_location = (0,0,0,0)

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {
                        'hide':FieldArg("boolean", False, jsonset=True),
                        'autofocus':FieldArg("boolean", True),
                        'action':FieldArg("url", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'button_text':FieldArg("text", 'Submit'),
                        'pin1':FieldArg("boolean", True, valdt=True),
                        'pin2':FieldArg("boolean", True, valdt=True),
                        'pin3':FieldArg("boolean", True, valdt=True),
                        'pin4':FieldArg("boolean", True, valdt=True),
                        'inner_div_class':FieldArg("cssclass", ''),
                        'char_div_class':FieldArg("cssclass", ''),
                        'char_div_style':FieldArg("cssstyle", ''),
                        'butt_div_class':FieldArg("cssclass", ''),
                        'button_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", ""),
                        'left_label':FieldArg("text", ''),
                        'left_class':FieldArg("cssclass", ''),
                        'left_style':FieldArg("cssstyle", ''),
                        'right_label':FieldArg("text", ''),
                        'right_class':FieldArg("cssclass", ''),
                        'right_style':FieldArg("cssstyle", '')
                       }


    def __init__(self, name=None, brief='', **field_args):
        """
        action: The page ident this button links to
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        button_text: The text on the button
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # inner div
        self[0] = tag.Part(tag_name="div")
        # error div
        self[0][0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0][0] = tag.Part(tag_name="p")
        self[0][0][0][0] = ''
        # container at 0,1
        self[0][1] = tag.Part(tag_name="div")
        self[0][1][0] = ""
        # form containing input text fields and button
        self[0][2] = tag.Part(tag_name="form", attribs={"role":"form", "method":"post"})
        # div containing text character fields
        self[0][2][0] = tag.Part(tag_name='div')
        # the first text input field
        self[0][2][0][0] = tag.ClosedPart(tag_name="input", attribs ={"type":"password",
                                                                      "size":"1",
                                                                      "maxlength":"1",
                                                                      "required":"required",
                                                                      "style":"width:2em;"})
        self[0][2][0][1] = "-"
        # the second text input field
        self[0][2][0][2] = tag.ClosedPart(tag_name="input", attribs ={"type":"password",
                                                                      "size":"1",
                                                                      "maxlength":"1",
                                                                      "required":"required",
                                                                      "style":"width:2em;"})
        self[0][2][0][3] = "-"
        # the third text input field
        self[0][2][0][4] = tag.ClosedPart(tag_name="input", attribs ={"type":"password",
                                                                      "size":"1",
                                                                      "maxlength":"1",
                                                                      "required":"required",
                                                                      "style":"width:2em;"})
        self[0][2][0][5] = "-"
        # the fourth text input field
        self[0][2][0][6] = tag.ClosedPart(tag_name="input", attribs ={"type":"password",
                                                                      "size":"1",
                                                                      "maxlength":"1",
                                                                      "required":"required",
                                                                      "style":"width:2em;"})
        # div containing submit button
        self[0][2][1] = tag.Part(tag_name='div')
        # the left label
        self[0][2][1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the submit button
        self[0][2][1][1] = tag.ClosedPart(tag_name="input", attribs = {"type":"submit"})
        # the right label
        self[0][2][1][2] = tag.Part(tag_name="label", hide_if_empty=True)



    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        if not self.get_field_value("action"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = skiboot.get_url(self.get_field_value("action"),  proj_ident=page.proj_ident)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return

        if self.get_field_value('inner_div_class'):
            self[0].update_attribs({"class":self.get_field_value('inner_div_class')})

        if self.get_field_value('error_class'):
            self[0][0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0][0].del_one_attrib("style")

        # update the action of the form
        self[0][2].update_attribs({"action": actionurl})

        if self.get_field_value('char_div_class'):
            self[0][2][0].update_attribs({"class":self.get_field_value('char_div_class')})
        if self.get_field_value('char_div_style'):
            self[0][2][0].update_attribs({"style":self.get_field_value('char_div_style')})

        # enable disable the pin number fields

        # first input field
        self[0][2][0][0].update_attribs({"name":self.get_formname('pin1')})
        if not self.get_field_value('pin1'):
            self[0][2][0][0].update_attribs({"disabled":"disabled", "value":"X"})

        # second input field
        self[0][2][0][2].update_attribs({"name":self.get_formname('pin2')})
        if not self.get_field_value('pin2'):
            self[0][2][0][2].update_attribs({"disabled":"disabled", "value":"X"})

        # third input field
        self[0][2][0][4].update_attribs({"name":self.get_formname('pin3')})
        if not self.get_field_value('pin3'):
            self[0][2][0][4].update_attribs({"disabled":"disabled", "value":"X"})

        # fourth input field
        self[0][2][0][6].update_attribs({"name":self.get_formname('pin4')})
        if not self.get_field_value('pin4'):
            self[0][2][0][6].update_attribs({"disabled":"disabled", "value":"X"})

        if self.get_field_value("autofocus"):
            # set autofocus on first non-disabled input field
            if self.get_field_value('pin1'):
                self[0][2][0][0].update_attribs({"autofocus":"autofocus"})
            elif self.get_field_value('pin2'):
                self[0][2][0][2].update_attribs({"autofocus":"autofocus"})
            elif self.get_field_value('pin3'):
                self[0][2][0][4].update_attribs({"autofocus":"autofocus"})
            elif self.get_field_value('pin4'):
                self[0][2][0][6].update_attribs({"autofocus":"autofocus"})

        # div containing the button
        if self.get_field_value('butt_div_class'):
            self[0][2][1].update_attribs({"class":self.get_field_value('butt_div_class')})

        if self.get_field_value('left_label'):
            self[0][2][1][0][0] = self.get_field_value('left_label')
        if self.get_field_value('left_class'):
            self[0][2][1][0].attribs = {"class": self.get_field_value('left_class')}
        if self.get_field_value('left_style'):
            self[0][2][1][0].update_attribs({"style": self.get_field_value('left_style')})
 
        # the submit button
        if self.get_field_value('button_class'):
            self[0][2][1][1].update_attribs({"class":self.get_field_value('button_class'),
                                                                                   "value":self.get_field_value('button_text')})
        else:
            self[0][2][1][1].update_attribs({"value":self.get_field_value('button_text')})

        # set an id in the submit button for the 'label for' tag
        self[0][2][1][1].insert_id()

        if self.get_field_value('right_label'):
            self[0][2][1][2][0] = self.get_field_value('right_label')
        if self.get_field_value('right_class'):
            self[0][2][1][2].attribs = {"class": self.get_field_value('right_class')}
        if self.get_field_value('right_style'):
            self[0][2][1][2].update_attribs({"style": self.get_field_value('right_style')})

        # set the label 'for' attribute
        self[0][2][1][0].update_attribs({'for':self[0][2][1][1].get_id()})
        self[0][2][1][2].update_attribs({'for':self[0][2][1][1].get_id()})

        # add ident and four hidden fields
        self.add_hiddens(self[0][2], page)



    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a keypress event handler"""
        jscript = """  $("#{ident} input").keypress(function (e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident = self.get_id())
        #if self._jsonurl:
        #    return jscript + self._make_fieldvalues('button_wait_text', url=self._jsonurl)
        #else:
        #    return jscript + self._make_fieldvalues('button_wait_text')
        return jscript


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div>  <!-- with class attribute set to inner_div_class if a class is set -->
    <div> <!-- normally hidden div, with class error_class -->
      <p> <!-- Any error text appears here --> </p>
    </div>
    <div>
     <!- A container of further code goes here ->
    </div>
    <form method="post"> <!-- action attribute set to action field -->
      <div>  <!-- with class attribute set to char_div_class and style to char_div_style -->
       <!-- each input field has maxlength set to one character -->
        <input type=\"password\" />-<input type=\"password\" />-<input type=\"password\" />-<input type=\"password\" />
      </div>
      <div>  <!-- with class attribute set to butt_div_class if a class is set -->
        <label> <!-- with class set to left_class and content to left_label -->
        </label>
        <input type=\"submit\" /> <!-- button value set to button_text -->
                                       <!-- and class attribute set to button_class if a class is set -->
        <label> <!-- with class set to right_class and content to right_label -->
        </label>
      </div>
      <!-- hidden input fields -->    
    </form>
  </div>
</div>"""



class NamePasswd1(Widget):
    """Defines a form containing text input and password fields."""

    error_location = (0,0,0)

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {'action':FieldArg("url", ''),
                        'action_json':FieldArg("url", ''),
                        'error_class':FieldArg("cssclass", ""),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'div1_class':FieldArg("cssclass", ''),
                        'label1':FieldArg("text", 'Username:'),
                        'label_class1':FieldArg("cssclass", ''),
                        'label_style1':FieldArg("cssstyle", ''),
                        'input_text1':FieldArg("text", '', valdt=True, jsonset=True),
                        'size1':FieldArg("text", ''),
                        'maxlength1':FieldArg("text", ''),
                        'disabled1':FieldArg("boolean", False),
                        'required1':FieldArg("boolean", True),
                        'pattern1':FieldArg("text", ''),
                        'title1':FieldArg("text", ''),
                        'set_input_accepted1':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored1':FieldArg("boolean", False, jsonset=True),
                        'div2_class':FieldArg("cssclass", ''),
                        'label2':FieldArg("text", 'Password:'),
                        'label_class2':FieldArg("cssclass", ''),
                        'label_style2':FieldArg("cssstyle", ''),
                        'input_text2':FieldArg("text", '', valdt=True, jsonset=True),
                        'size2':FieldArg("text", ''),
                        'maxlength2':FieldArg("text", ''),
                        'disabled2':FieldArg("boolean", False),
                        'required2':FieldArg("boolean", True),
                        'pattern2':FieldArg("text", ''),
                        'title2':FieldArg("text", ''),
                        'set_input_accepted2':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored2':FieldArg("boolean", False, jsonset=True),
                        'input_class1':FieldArg("cssclass", ''),
                        'input_class2':FieldArg("cssclass", ''),
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'div3_class':FieldArg("cssclass", ''),
                        'button_text':FieldArg("text", 'Submit'),
                        'button_class':FieldArg("cssclass", ''),
                        'button_style':FieldArg("cssstyle", ''),
                       }


    def __init__(self, name=None, brief='', **field_args):
        """
        action: The page ident this button links to
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        error_class: The class applied to the paragraph containing the error message on error.
        inputdiv_class: the class attribute of the tag which contains the label and inputs
        label1: The text displayed to the left of the text input field
        label_class1: The css class of the label
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        button_text: The text on the button
        button_class: the class attribute of the button
        input_accepted_class: A class which can be set on each input field
        input_errored_class: A class which can be set on each input field
        input_text1: The default text in the first text input field, field name used as the name attribute
        input_class1: Class set on the first input field
        size1: The number of characters appearing in the first text input area
        maxlength1: The maximum number of characters accepted in the first text area
        disabled1: Set True if the first input field is to be disabled
        required1: Set True to put the 'required' flag into the first input field
        pattern1: regular expression pattern for first input field
        title1: helps describe the pattern of the first input field
        label2: The text displayed to the left of the password input field
        label_class2: The css class of the label
        input_text2: The default text in the second text input field, field name used as the name attribute
        input_class2: Class set on the second input field
        size2: The number of characters appearing in the second text input area
        maxlength2: The maximum number of characters accepted in the second text area
        disabled2: Set True if the second field is to be disabled
        required2: Set True to put the 'required' flag into the second input field
        pattern2: regular expression pattern for the second input field
        title2: helps describe the pattern of the second input field
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})
        # div containing inputs and button
        self[1][0] = tag.Part(tag_name='div')
        # div containing label and input 1
        self[1][0][0] = tag.Part(tag_name='div')
        # the label 1
        self[1][0][0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the first text input field
        self[1][0][0][1] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})
        # div containing label and password input
        self[1][0][1] = tag.Part(tag_name='div')
        # the label 2
        self[1][0][1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the password field
        self[1][0][1][1] = tag.ClosedPart(tag_name="input", attribs ={"type":"password"})
        # div containing submit button
        self[1][0][2] = tag.Part(tag_name='div')
        # the submit button
        self[1][0][2][0] = tag.ClosedPart(tag_name="input", attribs={"type":"submit"})
        self._jsonurl = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        if not self.get_field_value("action"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = skiboot.get_url(self.get_field_value("action"),  proj_ident=page.proj_ident)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].update_attribs({"action": actionurl})
        self._jsonurl = skiboot.get_url(self.get_field_value("action_json"), proj_ident=page.proj_ident)
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")

        # the div holding labels, input fields and button
        if self.get_field_value('inputdiv_class'):
            self[1][0].attribs = {"class": self.get_field_value('inputdiv_class')}

        # div1
        if self.get_field_value('div1_class'):
            self[1][0][0].attribs = {"class": self.get_field_value('div1_class')}

        # set up label1
        if self.get_field_value('label1'):
            self[1][0][0][0][0] = self.get_field_value('label1')
        if self.get_field_value('label_class1'):
            self[1][0][0][0].attribs = {"class": self.get_field_value('label_class1')}
        if self.get_field_value('label_style1'):
            self[1][0][0][0].update_attribs({"style": self.get_field_value('label_style1')})

        # first input field
        self[1][0][0][1].update_attribs({"name":self.get_formname('input_text1'), "value":self.get_field_value('input_text1')})
        if self.get_field_value('size1'):
            self[1][0][0][1].update_attribs({"size":self.get_field_value('size1')})
        if self.get_field_value('maxlength1'):
            self[1][0][0][1].update_attribs({"maxlength":self.get_field_value('maxlength1')})
        if self.get_field_value('disabled1'):
            self[1][0][0][1].update_attribs({"disabled":"disabled"})
        if self.get_field_value('required1'):
            self[1][0][0][1].update_attribs({"required":"required"})
        if self.get_field_value('pattern1'):
            self[1][0][0][1].update_attribs({"pattern":self.get_field_value('pattern1')})
        if self.get_field_value('title1'):
            self[1][0][0][1].update_attribs({"title":self.get_field_value('title1')})

        if self.get_field_value('input_class1'):
            input_class1 = self.get_field_value('input_class1')
        else:
            input_class1 = ''

        if self.error_status and self.get_field_value('input_errored_class'):
            if input_class1:
                input_class1 = input_class1 + ' ' + self.get_field_value('input_errored_class')
            else:
                input_class1 = self.get_field_value('input_errored_class')
        elif self.get_field_value('set_input_errored1') and self.get_field_value('input_errored_class'):
            if input_class1:
                input_class1 = input_class1 + ' ' + self.get_field_value('input_errored_class')
            else:
                input_class1 = self.get_field_value('input_errored_class')
        elif self.get_field_value('set_input_accepted1') and self.get_field_value('input_accepted_class'):
            if input_class1:
                input_class1 = input_class1 + ' ' + self.get_field_value('input_accepted_class')
            else:
                input_class1 = self.get_field_value('input_accepted_class')

        if input_class1:
            self[1][0][0][1].update_attribs({"class":input_class1})

        # set up label2
        if self.get_field_value('label2'):
            self[1][0][1][0][0] = self.get_field_value('label2')
        if self.get_field_value('label_class2'):
            self[1][0][1][0].attribs = {"class": self.get_field_value('label_class2')}
        if self.get_field_value('label_style2'):
            self[1][0][1][0].update_attribs({"style": self.get_field_value('label_style2')})

        # password field
        self[1][0][1][1].update_attribs({"name":self.get_formname('input_text2'), "value":self.get_field_value('input_text2')})
        if self.get_field_value('size2'):
            self[1][0][1][1].update_attribs({"size":self.get_field_value('size2')})
        if self.get_field_value('maxlength2'):
            self[1][0][1][1].update_attribs({"maxlength":self.get_field_value('maxlength2')})
        if self.get_field_value('disabled2'):
            self[1][0][1][1].update_attribs({"disabled":"disabled"})
        if self.get_field_value('required2'):
            self[1][0][1][1].update_attribs({"required":"required"})
        if self.get_field_value('pattern2'):
            self[1][0][1][1].update_attribs({"pattern":self.get_field_value('pattern2')})
        if self.get_field_value('title2'):
            self[1][0][1][1].update_attribs({"title":self.get_field_value('title2')})

        if self.get_field_value('input_class2'):
            input_class2 = self.get_field_value('input_class2')
        else:
            input_class2 = ''

        if self.error_status and self.get_field_value('input_errored_class'):
            if input_class2:
                input_class2 = input_class2 + ' ' + self.get_field_value('input_errored_class')
            else:
                input_class2 = self.get_field_value('input_errored_class')
        elif self.get_field_value('set_input_errored2') and self.get_field_value('input_errored_class'):
            if input_class2:
                input_class2 = input_class2 + ' ' + self.get_field_value('input_errored_class')
            else:
                input_class2 = self.get_field_value('input_errored_class')
        elif self.get_field_value('set_input_accepted2') and self.get_field_value('input_accepted_class'):
            if input_class2:
                input_class2 = input_class2 + ' ' + self.get_field_value('input_accepted_class')
            else:
                input_class2 = self.get_field_value('input_accepted_class')

        if input_class2:
            self[1][0][1][1].update_attribs({"class":input_class2})

        # div3
        if self.get_field_value('div3_class'):
            self[1][0][2].attribs = {"class": self.get_field_value('div3_class')}
        # the submit button
        self[1][0][2][0].update_attribs({"value":self.get_field_value('button_text')})
        # the button class
        if self.get_field_value('button_class'):
            self[1][0][2][0].update_attribs({"class": self.get_field_value('button_class')})
        if self.get_field_value('button_style'):
            self[1][0][2][0].update_attribs({"style": self.get_field_value('button_style')})


        # set an id in the first input field for the 'label for' tag
        self[1][0][0][1].insert_id()
        # set the label 'for' attribute
        self[1][0][0][0].update_attribs({'for':self[1][0][0][1].get_id()})

        # set an id in the password field for the 'label for' tag
        self[1][0][1][1].insert_id()
        # set the label 'for' attribute
        self[1][0][1][0].update_attribs({'for':self[1][0][1][1].get_id()})

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        jscript = """  $("#{ident} form").on("submit input", function(e) {{
    SKIPOLE.widgets["{ident}"].eventfunc(e);
    }});
""".format(ident=self.get_id())
        if self._jsonurl:
            return jscript + self._make_fieldvalues('input_accepted_class', 'input_errored_class', url=self._jsonurl)
        return jscript + self._make_fieldvalues('input_accepted_class', 'input_errored_class')


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <form method="post"> <!-- action attribute set to action field -->
    <div> <!-- class attribute set to inputdiv_class -->
      <div> <!-- class attribute set to div1_class -->
        <label> <!-- with class set to label_class1 and content to label1 -->
        </label>
        <input type=\"text\" /> <!-- input text value set to input_text1, class to input_class1 -->
      </div>
      <div> <!-- class attribute set to div2_class -->
        <label> <!-- with class set to label_class2 and content to label2 -->
        </label>
        <input type=\"password\" /> <!-- input text value set to input_text2, class to input_class2 -->
      </div>
      <div> <!-- class attribute set to div3_class -->
        <input type=\"submit\" /> <!-- button value set to button_text and class to button_class -->
      </div>
    </div>
    <!-- hidden input fields -->    
  </form>
</div>"""












"""Functions for editing a section"""

import re

from collections import OrderedDict

from ..ski import skiboot, tag, read_json
from ..ski.excepts import ServerError

from . import project_loaded, get_proj_section, get_proj_page, insert_item_in_page, del_location_in_section, insert_item_in_section

from .info_tuple import PlaceHolderInfo, SectionElement, SectionTextBlock


# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')



def list_section_names(project=None):
    """Returns a list of section names in the project"""
    if project is None:
        project = skiboot.project_ident()
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        return ()
    return proj.list_section_names()


def placeholder_info(project, pagenumber, location):
    """location is a tuple or list consisting of three items:
       a string (such as head or widget name)
       a container integer, such as 0 for widget container 0, or None if not in container
       a tuple or list of location integers
       returns None if placeholder not found, otherwise returns a namedtuple with items
       project, pagenumber, section_name, alias, brief, multiplier, mtag
    """
    proj, page = get_proj_page(project, pagenumber)
    part = page.location_item(location)
    if part is None:
        return

    if part.__class__.__name__ != "SectionPlaceHolder":
        return

    if hasattr(part, 'brief'):
        brief = part.brief
    else:
        brief = None

    if hasattr(part, 'section_name'):
        section_name = part.section_name
    else:
        section_name = None

    if hasattr(part, 'placename'):
        alias = part.placename
    else:
        alias = None

    if hasattr(part, 'multiplier'):
        multiplier = part.multiplier
    else:
        multiplier = None

    if hasattr(part, 'mtag'):
        mtag = part.mtag
    else:
        mtag = None

    return PlaceHolderInfo(project, pagenumber, section_name, alias, brief, multiplier, mtag)


def edit_placeholder(project, pagenumber, pchange, location, section_name, alias, brief, multiplier, mtag):
    """Given a placeholder at project, pagenumber, location
       sets the values section_name, alias, brief, multiplier, mtag, returns page change uuid """
    # raise error if invalid project
    if project is None:
        project = skiboot.project_ident()
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError(message="The edited project is invalid")

    location_string, container_number, location_list = location

    part = None

    if pagenumber is None:
        raise ServerError(message="The page containing this section seems to be invalid")

    ident = skiboot.find_ident(pagenumber, project)
    if not ident:
        raise ServerError(message="The page containing this section seems to be invalid")
    page = skiboot.get_item(ident)
    if not page:
        raise ServerError(message="The page containing this section seems to be invalid")
    if (page.page_type != "TemplatePage") and (page.page_type != "SVG"):
        raise ServerError(message="The page type containing this section seems to be invalid")
    if page.change != pchange:
        raise ServerError(message="The page has been changed prior to this submission, someone else may be editing this project")
    part = page.location_item(location)
    if part is None:
        raise ServerError(message="Cannot find the section placeholder")

    if part.__class__.__name__ != "SectionPlaceHolder":
        raise ServerError(message="The edited location does not seem to be a section placeholder")

    # is placename unique
    if alias != part.placename:
        if alias in page.widgets:
            raise ServerError(message="Duplicate name in the page - a widget has this assigned")
        if alias in page.section_places:
            raise ServerError(message="This name clashes with a section alias already within this page")
    part.section_name = section_name
    part.placename = alias
    part.brief = brief
    part.multiplier = multiplier
    part.mtag = mtag

    # save the altered page, and return the page.change uuid
    return proj.save_page(page)



def new_placeholder(project, pagenumber, pchange, location, section_name, alias, brief):
    """Create a new placeholder at project, pagenumber, location
       with section_name, alias, brief returns page change uuid and new location"""
    # create new placeholder
    newplaceholder = tag.SectionPlaceHolder(section_name=section_name,
                                            placename=alias,
                                            brief=brief)
    # call skilift.insert_item_in_page to insert the item, save the page and return pchange and new location
    return insert_item_in_page(project, pagenumber, pchange, location, newplaceholder)


def sectionchange(project, section_name):
    """Returns the section change

       Given a project name, and an existing section name within the project,
       returns a 'change' uuid associated  with the section. A new random uuid
       is generated every time a section is edited, and is therefore a measure
       that a given section both exists, and has not changed since the last time
       it was accessed.
       The project must be currently loaded as either the root project or a sub-project.
       Returns None if no section with this name is found."""
    # raise error if invalid project
    project_loaded(project)
    if not isinstance(section_name, str):
         raise ServerError(message="Given section_name is invalid")
    proj = skiboot.getproject(project)
    section = proj.section(section_name, makecopy=False)
    if section is None:
        return
    return section.change


def del_location(project, section_name, schange, location):
    "Deletes the item at the given location in the section, returns new schange"
    return del_location_in_section(project, section_name, schange, location)


def move_location(project, section_name, schange, from_location, to_location):
    """Move an item in the given section from one spot to another, defined by its location
       Returns new section change uuid"""
    # raise error if invalid project
    proj, section = get_proj_section(project, section_name, schange)

    if to_location == from_location:
        # no movement
        return schange

    from_location_string, from_container, from_location_integers = from_location
    to_location_string, to_container, to_location_integers = to_location

    if (from_location_string != to_location_string) or (from_container != to_container):
        raise ServerError(message="Unable to move")

    if from_container is None:
        if from_location_string != section_name:
            raise ServerError(message="Unable to move item")
        # location integers are integers within the section, move it 
        up = False
        i = 0
        while True:
            if to_location_integers[i] < from_location_integers[i]:
                up = True
                break
            if to_location_integers[i] > from_location_integers[i]:
                # up is False
                break
            # so this digit is the same
            i += 1
            if len(from_location_integers) == i:
                # from location integers exhausted and equal to to_location! Trying to move an item under itself
                return schange
            if len(to_location_integers) == i:
                up = True
                break
        if up:
            # move in the upwards direction
            #    delete it from original location
            #    insert it into new location
            try:
                # get the part at from_location_integers
                part = section.get_location_value(from_location_integers)
                # delete part from current location
                section.del_location_value(from_location_integers)
                # and insert it in the new location
                section.insert_location_value(to_location_integers, part)
            except Exception:
                raise ServerError(message="Unable to move item")
        else:
            # move in the downwards direction
            #    insert it into new location
            #    delete it from original location
            try:
                # get the part at from_location_integers
                part = section.get_location_value(from_location_integers)
                # and insert it in the new location
                section.insert_location_value(to_location_integers, part)
                # delete part from current location
                section.del_location_value(from_location_integers)
            except Exception:
                raise ServerError(message="Unable to move item")
        # And save this section copy to the project
        return proj.add_section(section_name, section)

    # part is within a widget, so get its location relative to the section
    widget = section.widgets[from_location_string]
    ident_string = widget.ident_string

    # ident_string is sectionname-x-y

    splitstring = ident_string.split("-")
    widg_ints = [ int(i) for i in splitstring[1:] ]

    widg_container_ints = list(widget.get_container_loc(from_container))

    from_location_ints = widg_ints + widg_container_ints + list(from_location_integers)
    to_location_ints = widg_ints + widg_container_ints + list(to_location_integers)

    # and move this item by calling this function again
    return move_location(project, section_name, schange, (section_name, None, from_location_ints), (section_name, None, to_location_ints))


def create_new_section(project, section_name, tag_name, brief):
    "Create a new section, return new section change on success, raises a ServerError on failure"
    # raise error if invalid project
    if not section_name:
        raise ServerError(message = "Section name missing")
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError(message="The edited project is invalid")
    section_list = proj.list_section_names()
    if section_name in section_list:
        raise ServerError(message = "Section name already exists")
    section_lower_name = section_name.lower()
    if (section_lower_name == 'body') or (section_lower_name == 'head') or (section_lower_name == 'svg'):
        raise ServerError(message="Unable to create the section, the name given is reserved")
    if _AN.search(section_name):
        raise ServerError(message="Invalid section name, alphanumeric and underscore only")
    if section_name[0] == '_':
        raise ServerError(message="Invalid section name, must not start with an underscore")
    if section_name.isdigit():
        raise ServerError(message="Unable to create the section, the name must include some letters")
    # creates and adds the section
    return proj.add_section(section_name, tag.Section(tag_name=tag_name, brief=brief))


def delete_section(project, section_name):
    "delete this section, return None on success, raises a ServerError on failure"
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    if proj is None:
        raise ServerError(message="The edited project is invalid")
    section_list = proj.list_section_names()
    if section_name not in section_list:
        raise ServerError(message="Section name does not exists")
    proj.delete_section(section_name)


def create_html_comment_in_section(project, section_name, schange, location, text="comment here"):
    "Creates a new html comment in the given section, returns the new schange and comment location"
    com = tag.Comment(text)
    # call skilift.insert_item_in_section to insert the item, save the section and return schange
    new_schange, new_location = insert_item_in_section(project, section_name, schange, location, com)
    return new_schange, new_location


def create_html_symbol_in_section(project, section_name, schange, location, text="&nbsp;"):
    "Creates a new html symbol in the given section, returns the new schange and symbol location"
    sym = tag.HTMLSymbol(text)
    # call skilift.insert_item_in_section to insert the item, save the section and return schange
    new_schange, new_location = insert_item_in_section(project, section_name, schange, location, sym)
    return new_schange, new_location


def create_html_element_in_section(project, section_name, schange, location, name, brief, opentag = True):
    "Creates a new html element in the given section, returns the new schange"
    if opentag:
        newpart = tag.Part(tag_name=name, brief=brief)
    else:
        newpart = tag.ClosedPart(tag_name=name, brief=brief)
    # call skilift.insert_item_in_section to insert the item, save the section and return schange
    new_schange, new_location = insert_item_in_section(project, section_name, schange, location, newpart)
    return new_schange


def create_part_in_section(project, section_name, schange, location, json_data):
    """Builds the part from the given json string or ordered dictionary, and adds it to project either inserted into the html element
       currently at the given part location, or if not an element that can accept contents, inserted after the element.
       Returns new schange value"""
    proj, section = get_proj_section(project, section_name, schange)
    try:
        newpart = read_json.make_part_for_section(section, json_data)
    except Exception:
        raise ServerError("Unable to create part")
    # call skilift.insert_item_in_section to insert the item, save the section and return schange
    new_schange, new_location = insert_item_in_section(project, section_name, schange, location, newpart)
    return new_schange


def create_item_in_section(project, section_name, schange, location, json_string):
    """Builds the item from the given json string, and adds it to project either inserted into the html element
       currently at the given location, or if not an element that can accept contents, inserted after the element.
       Returns new schange value"""
    proj, section = get_proj_section(project, section_name, schange)
    try:
        newitem = read_json.make_item_for_section(section, json_string)
    except Exception:
        raise ServerError("Unable to create item")
    # call skilift.insert_item_in_section to insert the item, save the section and return schange
    new_schange, new_location = insert_item_in_section(project, section_name, schange, location, newitem)
    return new_schange



def section_element(project, section_name, schange, location):
    """Return a SectionElement tuple for the html element at the given location"""

    proj, section = get_proj_section(project, section_name, schange)
    part = section.location_item(location)

    if part is None:
        raise ServerError("The item at the location has not been found")

    if hasattr(part, '__class__'):
        part_type = part.__class__.__name__
    else:
        raise ServerError("The item at the location must be an html element")

    if (part_type != "Part") and (part_type != "ClosedPart"):
        # part_type is normally a Part or ClosedPart, but may also be a Section
        if (part_type == "Section"):
            # however, in the special case where it is a section, there must be no location digits
            # a section cannot contain another section
            if location[2]:
                raise ServerError("The item at the location cannot be a Section!")
        else:
            # Not a section either
            raise ServerError("The item at the location must be an html element")

    tag_name = part.tag_name
    brief = part.brief
    show = part.show
    hide_if_empty = part.hide_if_empty
    attribs = OrderedDict(sorted(part.attribs.items(), key=lambda t: t[0]))

    return SectionElement(project, section_name, schange, location, part_type, tag_name, brief, show, hide_if_empty, attribs)


def edit_section_element(project, section_name, schange, location, tag_name, brief, hide_if_empty, attribs):
    """Given an element at project, section_name, location
       sets the element values, returns section change uuid """
    proj, section = get_proj_section(project, section_name, schange)
    part = section.location_item(location)
    part.tag_name = tag_name
    part.brief = brief
    part.attribs = OrderedDict(sorted(attribs.items(), key=lambda t: t[0]))
    if part.__class__.__name__ == "Part":
        part.hide_if_empty = hide_if_empty
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def del_attrib(project, section_name, schange, location, attribute):
    """Given an element at project, section_name, location
       deletes the given attribute, returns sceetion change uuid """
    proj, section = get_proj_section(project, section_name, schange)
    part = section.location_item(location)
    part.del_one_attrib(attribute)
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def get_text(project, section_name, schange, location):
    """Return a text string from the page at the given location"""
    proj, section = get_proj_section(project, section_name, schange)
    text = section.location_item(location)
    if not isinstance(text, str):
        raise ServerError("Item at this location is not identified as a text string")
    return text


def get_symbol(project, section_name, schange, location):
    """Return a symbol from the section at the given location"""
    proj, section = get_proj_section(project, section_name, schange)
    sym = section.location_item(location)
    if not isinstance(sym, tag.HTMLSymbol):
        raise ServerError("Item at this location is not identified as a HTML Symbol")
    return sym.text


def edit_section_symbol(project, section_name, schange, location, text):
    """Given a symbol at project, section_name, location
       sets the symbol, returns section change uuid """
    proj, section = get_proj_section(project, section_name, schange)
    sym = section.location_item(location)
    if not isinstance(sym, tag.HTMLSymbol):
        raise ServerError("Item at this location is not identified as a HTML Symbol")
    sym.text = text
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def get_comment(project, section_name, schange, location):
    """Return a comment from the section at the given location"""
    proj, section = get_proj_section(project, section_name, schange)
    com = section.location_item(location)
    if not isinstance(com, tag.Comment):
        raise ServerError("Item at this location is not identified as a HTML comment")
    return com.text


def edit_section_comment(project, section_name, schange, location, text):
    """Given a comment at project, section_name, location
       sets the comment, returns section change uuid """
    proj, section = get_proj_section(project, section_name, schange)
    com = section.location_item(location)
    if not isinstance(com, tag.Comment):
        raise ServerError("Item at this location is not identified as a HTML comment")
    com.text = text
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)
 

def section_textblock(project, section_name, schange, location):
    """Return a SectionTextBlock tuple for the textblock at the given location"""
    proj, section = get_proj_section(project, section_name, schange)
    tblock = section.location_item(location)
    if not isinstance(tblock, tag.TextBlock):
        raise ServerError("Item at this location is not identified as a TextBlock")
    return SectionTextBlock(project, section_name, schange, location, tblock.textref, tblock.project, tblock.failmessage, tblock.escape, tblock.linebreaks)


def create_textblock_in_section(project, section_name, schange, location, textref, failmessage, escape, linebreaks, tblock_project=''):
    "Creates a new textblock in the given section, returns the new schange and location"
    tblock = tag.TextBlock(textref=textref, project=tblock_project, failmessage=failmessage, escape=escape, linebreaks=linebreaks)
    # call skilift.insert_item_in_section to insert the item, save the section and return schange
    new_schange, new_location = insert_item_in_section(project, section_name, schange, location, tblock)
    return new_schange, new_location


def edit_section_textblock(project, section_name, schange, location, textref, tblock_project, failmessage, escape, linebreaks):
    """Given an textblock at project, section_name, location
       sets the textblock values, returns section change uuid """
    proj, section = get_proj_section(project, section_name, schange)
    tblock = section.location_item(location)
    if not isinstance(tblock, tag.TextBlock):
        raise ServerError("Item at this location is not identified as a TextBlock")
    tblock.textref = textref
    tblock.failmessage = failmessage
    tblock.escape = escape
    tblock.linebreaks = linebreaks
    if project == tblock_project:
        tblock.project = ''
    else:
        tblock.project = tblock_project
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)

